"""gr.CheckboxGroup() component"""

from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, Literal

from gradio_client.documentation import document

from gradio.components.base import Component, FormComponent
from gradio.components.button import Button
from gradio.events import Events
from gradio.exceptions import Error
from gradio.i18n import I18nData
from gradio.utils import set_default_buttons

if TYPE_CHECKING:
    from gradio.components import Timer


@document()
class CheckboxGroup(FormComponent):
    """
    Creates a set of checkboxes. Can be used as an input to pass a set of values to a function or as an output to display values, a subset of which are selected.
    Demos: sentence_builder
    """

    EVENTS = [Events.change, Events.input, Events.select]

    def __init__(
        self,
        choices: Sequence[str | int | float | tuple[str, str | int | float]]
        | None = None,
        *,
        value: Sequence[str | float | int] | str | float | int | Callable | None = None,
        type: Literal["value", "index"] = "value",
        label: str | I18nData | None = None,
        info: str | I18nData | None = None,
        every: Timer | float | None = None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        show_label: bool | None = None,
        show_select_all: bool = False,
        container: bool = True,
        scale: int | None = None,
        min_width: int = 160,
        interactive: bool | None = None,
        visible: bool | Literal["hidden"] = True,
        elem_id: str | None = None,
        elem_classes: list[str] | str | None = None,
        render: bool = True,
        key: int | str | tuple[int | str, ...] | None = None,
        preserved_by_key: list[str] | str | None = "value",
        buttons: list[Button] | None = None,
    ):
        """
        Parameters:
            choices: A list of string or numeric options to select from. An option can also be a tuple of the form (name, value), where name is the displayed name of the checkbox button and value is the value to be passed to the function, or returned by the function.
            value: Default selected list of options. If a single choice is selected, it can be passed in as a string or numeric type. If a function is provided, the function will be called each time the app loads to set the initial value of this component.
            type: Type of value to be returned by component. "value" returns the list of strings of the choices selected, "index" returns the list of indices of the choices selected.
            label: the label for this component, displayed above the component if `show_label` is `True` and is also used as the header if there are a table of examples for this component. If None and used in a `gr.Interface`, the label will be the name of the parameter this component corresponds to.
            info: additional component description, appears below the label in smaller font. Supports markdown / HTML syntax.
            every: Continously calls `value` to recalculate it if `value` is a function (has no effect otherwise). Can provide a Timer whose tick resets `value`, or a float that provides the regular interval for the reset Timer.
            inputs: Components that are used as inputs to calculate `value` if `value` is a function (has no effect otherwise). `value` is recalculated any time the inputs change.
            show_label: If True, will display label.
            show_select_all: If True, will display a select/deselect all checkbox next to the label. Only available when show_label is True.
            container: If True, will place the component in a container - providing some extra padding around the border.
            scale: Relative width compared to adjacent Components in a Row. For example, if Component A has scale=2, and Component B has scale=1, A will be twice as wide as B. Should be an integer.
            min_width: Minimum pixel width, will wrap if not sufficient screen space to satisfy this value. If a certain scale value results in this Component being narrower than min_width, the min_width parameter will be respected first.
            interactive: If True, choices in this checkbox group will be checkable; if False, checking will be disabled. If not provided, this is inferred based on whether the component is used as an input or output.
            visible: If False, component will be hidden. If "hidden", component will be visually hidden and not take up space in the layout but still exist in the DOM
            elem_id: An optional string that is assigned as the id of this component in the HTML DOM. Can be used for targeting CSS styles.
            elem_classes: An optional list of strings that are assigned as the classes of this component in the HTML DOM. Can be used for targeting CSS styles.
            render: If False, component will not render be rendered in the Blocks context. Should be used if the intention is to assign event listeners now but render the component later.
            key: in a gr.render, Components with the same key across re-renders are treated as the same component, not a new component. Properties set in 'preserved_by_key' are not reset across a re-render.
            preserved_by_key: A list of parameters from this component's constructor. Inside a gr.render() function, if a component is re-rendered with the same key, these (and only these) parameters will be preserved in the UI (if they have been changed by the user or an event listener) instead of re-rendered based on the values provided during constructor.
            buttons: A list of gr.Button() instances to show in the top right corner of the component. Custom buttons will appear in the toolbar with their configured icon and/or label, and clicking them will trigger any .click() events registered on the button.
        """
        self.choices = (
            # Although we expect choices to be a list of tuples, it can be a list of tuples if the Gradio app
            # is loaded with gr.load() since Python tuples are converted to lists in JSON.
            [tuple(c) if isinstance(c, (tuple, list)) else (str(c), c) for c in choices]
            if choices
            else []
        )
        valid_types = ["value", "index"]
        if type not in valid_types:
            raise ValueError(
                f"Invalid value for parameter `type`: {type}. Please choose from one of: {valid_types}"
            )
        self.type = type
        self.show_select_all = show_select_all
        super().__init__(
            label=label,
            info=info,
            every=every,
            inputs=inputs,
            show_label=show_label,
            container=container,
            scale=scale,
            min_width=min_width,
            interactive=interactive,
            visible=visible,
            elem_id=elem_id,
            elem_classes=elem_classes,
            render=render,
            key=key,
            preserved_by_key=preserved_by_key,
            value=value,
        )
        self.buttons = set_default_buttons(buttons, None)
        self._value_description = f"one or more of {self.choices}"

    def example_payload(self) -> Any:
        return [self.choices[0][1]] if self.choices else None

    def example_value(self) -> Any:
        return [self.choices[0][1]] if self.choices else None

    def api_info(self) -> dict[str, Any]:
        return {
            "items": {"enum": [c[1] for c in self.choices], "type": "string"},
            "title": "Checkbox Group",
            "type": "array",
        }

    def preprocess(
        self, payload: list[str | int | float]
    ) -> list[str | int | float] | list[int | None]:
        """
        Parameters:
            payload: the list of checked checkboxes' values
        Returns:
            Passes the list of checked checkboxes as a `list[str | int | float]` or their indices as a `list[int]` into the function, depending on `type`.
        """
        choice_values = [value for _, value in self.choices]
        for value in payload:
            if value not in choice_values:
                raise Error(
                    f"Value: {value!r} (type: {type(value)}) is not in the list of choices: {choice_values}"
                )
        if self.type == "value":
            return payload
        elif self.type == "index":
            return [
                choice_values.index(choice) if choice in choice_values else None
                for choice in payload
            ]
        else:
            raise ValueError(
                f"Unknown type: {self.type}. Please choose from: 'value', 'index'."
            )

    def postprocess(
        self, value: list[str | int | float] | str | int | float | None
    ) -> list[str | int | float]:
        """
        Parameters:
            value: Expects a `list[str | int | float]` of values or a single `str | int | float` value, the checkboxes with these values are checked.
        Returns:
            the list of checked checkboxes' values
        """
        if value is None:
            return []
        if not isinstance(value, list):
            value = [value]
        return value



import io
import sys
from contextlib import contextmanager
from functools import partial
from string import capwords
from unittest.mock import MagicMock, patch

import httpx
import pytest

import gradio
from gradio.blocks import Blocks
from gradio.components import Image, Textbox
from gradio.interface import Interface, TabbedInterface, close_all, os
from gradio.layouts import TabItem, Tabs
from gradio.utils import assert_configs_are_equivalent_besides_ids


@contextmanager
def captured_output():
    new_out, new_err = io.StringIO(), io.StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class TestInterface:
    @pytest.mark.serial
    def test_close(self):
        io = Interface(lambda input: None, "textbox", "label")
        _, local_url, _ = io.launch(prevent_thread_lock=True)
        response = httpx.get(local_url)
        assert response.status_code == 200
        io.close()
        with pytest.raises(Exception):
            response = httpx.get(local_url)

    def test_close_all(self):
        interface = Interface(lambda input: None, "textbox", "label")
        interface.close = MagicMock()  # type: ignore
        close_all()
        interface.close.assert_not_called()  # type: ignore
        interface.launch(prevent_thread_lock=True)
        close_all()
        interface.close.assert_called()  # type: ignore

    def test_no_input_or_output(self):
        with pytest.raises(TypeError):
            Interface(lambda x: x, examples=1234)  # type: ignore

    def test_partial_functions(self):
        def greet(name, formatter):
            return formatter(f"Hello {name}!")

        greet_upper_case = partial(greet, formatter=capwords)
        demo = Interface(fn=greet_upper_case, inputs="text", outputs="text")
        assert demo("abubakar") == "Hello Abubakar!"

    def test_input_labels_extracted_from_method(self):
        class A:
            def test(self, parameter_name):
                return parameter_name

        t = Textbox()
        Interface(A().test, t, "text")
        assert t.label == "parameter_name"

        def test(parameter_name1, parameter_name2):
            return parameter_name1

        t = Textbox()
        i = Image()
        Interface(test, (t, i), "text")
        assert t.label == "parameter_name1"
        assert i.label == "parameter_name2"

        def special_args_test(req: gradio.Request, parameter_name):
            return parameter_name

        t = Textbox()
        Interface(special_args_test, t, "text")
        assert t.label == "parameter_name"

    def test_examples_valid_path(self):
        path = os.path.join(
            os.path.dirname(__file__), "../gradio/test_data/flagged_with_log"
        )
        interface = Interface(lambda x: 3 * x, "number", "number", examples=path)
        dataset_check = any(
            c["type"] == "dataset" for c in interface.get_config_file()["components"]
        )
        assert dataset_check

    @patch("time.sleep")
    @pytest.mark.serial
    def test_block_thread(self, mock_sleep):
        with pytest.raises(KeyboardInterrupt):
            with captured_output() as (out, _):
                mock_sleep.side_effect = KeyboardInterrupt()
                interface = Interface(lambda x: x, "textbox", "label")
                interface.launch(prevent_thread_lock=False)
                output = out.getvalue().strip()
                assert (
                    "Keyboard interruption in main thread... closing server." in output
                )

    @patch("gradio.utils.colab_check")
    @patch("gradio.networking.setup_tunnel")
    def test_launch_colab_share_error(self, mock_setup_tunnel, mock_colab_check):
        mock_setup_tunnel.side_effect = RuntimeError()
        mock_colab_check.return_value = True
        interface = Interface(lambda x: x, "textbox", "label")
        _, _, share_url = interface.launch(prevent_thread_lock=True)
        assert share_url is None
        interface.close()

    def test_interface_representation(self):
        def prediction_fn(x):
            return x

        prediction_fn.__name__ = "prediction_fn"
        repr = str(Interface(prediction_fn, "textbox", "label")).split("\n")
        assert prediction_fn.__name__ in repr[0]
        assert len(repr[0]) == len(repr[1])

    @patch("webbrowser.open")
    def test_interface_browser(self, mock_browser):
        interface = Interface(lambda x: x, "textbox", "label")
        interface.launch(inbrowser=True, prevent_thread_lock=True)
        mock_browser.assert_called_once()
        interface.close()

    def test_examples_list(self):
        examples = ["test1", "test2"]
        interface = Interface(
            lambda x: x, "textbox", "label", examples=examples, examples_per_page=2
        )
        interface.launch(prevent_thread_lock=True)
        assert len(interface.examples_handler.examples) == 2
        assert len(interface.examples_handler.examples[0]) == 1
        assert interface.examples_handler.dataset.get_config()["samples_per_page"] == 2
        interface.close()

    @patch("IPython.display.display")
    def test_inline_display(self, mock_display):
        interface = Interface(lambda x: x, "textbox", "label")
        interface.launch(inline=True, prevent_thread_lock=True)
        mock_display.assert_called_once()
        interface.launch(inline=True, prevent_thread_lock=True)
        assert mock_display.call_count == 2
        interface.close()

    def test_setting_interactive_false(self):
        output_textbox = Textbox()
        Interface(lambda x: x, "textbox", output_textbox)
        assert not output_textbox.get_config()["interactive"]
        output_textbox = Textbox(interactive=True)
        Interface(lambda x: x, "textbox", output_textbox)
        assert output_textbox.get_config()["interactive"]

    def test_get_api_info(self):
        io = Interface(lambda x: x, Image(type="filepath"), "textbox")
        api_info = io.get_api_info()
        assert api_info
        assert len(api_info["named_endpoints"]) == 1
        assert len(api_info["unnamed_endpoints"]) == 0

    def test_api_name(self):
        io = Interface(lambda x: x, "textbox", "textbox", api_name="echo")
        assert next(
            (d for d in io.config["dependencies"] if d["api_name"] == "echo"),  # type: ignore
            None,
        )

    def test_show_progress(self):
        io = Interface(
            lambda x: x, "textbox", "textbox", api_name="echo", show_progress="hidden"
        )
        dependency = next(
            (d for d in io.config["dependencies"] if d["api_name"] == "echo"),  # type: ignore
            None,
        )
        assert dependency and dependency["show_progress"] == "hidden"

    def test_interface_in_blocks_does_not_error(self):
        with Blocks():
            Interface(fn=lambda x: x, inputs=Textbox(), outputs=Image())

    def test_interface_with_built_ins(self):
        t = Textbox()
        Interface(fn=str, inputs=t, outputs=Textbox())
        assert t.label == "input 0"

    def test_interface_additional_components_are_included_as_inputs(self):
        t = Textbox()
        s = gradio.Slider(0, 100)
        io = Interface(fn=str, inputs=t, outputs=Textbox(), additional_inputs=s)
        assert io.input_components == [t, s]


class TestTabbedInterface:
    def test_tabbed_interface_config_matches_manual_tab(self):
        interface1 = Interface(lambda x: x, "textbox", "textbox")
        interface2 = Interface(lambda x: x, "image", "image")

        with Blocks(mode="tabbed_interface") as demo:
            with Tabs():
                with TabItem(label="tab1"):
                    interface1.render()
                with TabItem(label="tab2"):
                    interface2.render()

        interface3 = Interface(lambda x: x, "textbox", "textbox")
        interface4 = Interface(lambda x: x, "image", "image")
        tabbed_interface = TabbedInterface([interface3, interface4], ["tab1", "tab2"])

        assert assert_configs_are_equivalent_besides_ids(
            demo.get_config_file(),  # type: ignore
            tabbed_interface.get_config_file(),  # type: ignore
        )


@pytest.mark.parametrize(
    "interface_type", ["standard", "input_only", "output_only", "unified"]
)
@pytest.mark.parametrize("live", [True, False])
@pytest.mark.parametrize("use_generator", [True, False])
def test_interface_adds_stop_button(interface_type, live, use_generator):
    def gen_func(inp):
        yield inp

    def func(inp):
        return inp

    if interface_type == "standard":
        interface = gradio.Interface(
            gen_func if use_generator else func, "number", "number", live=live
        )
    elif interface_type == "input_only":
        interface = gradio.Interface(
            gen_func if use_generator else func, "number", None, live=live
        )
    elif interface_type == "output_only":
        interface = gradio.Interface(
            gen_func if use_generator else func, None, "number", live=live
        )
    else:
        num = gradio.Number()
        interface = gradio.Interface(
            gen_func if use_generator else func, num, num, live=live
        )
    has_stop = (
        len(
            [
                c
                for c in interface.config["components"]
                if c["props"].get("variant", "") == "stop"
            ]
        )
        == 1
    )
    if use_generator and not live:
        assert has_stop
    else:
        assert not has_stop


def test_live_interface_sets_always_last():
    iface = gradio.Interface(
        fn=lambda s: s,
        inputs=gradio.Textbox(lines=2, placeholder="Hello 👋", label="Input Sentence"),
        outputs=gradio.Markdown(),
        live=True,  # Set live to True for real-time feedback
    )
    config = iface.get_config_file()
    assert "dependencies" in config
    for dep in config["dependencies"]:
        if dep["targets"][0][1] == "change":
            assert dep["trigger_mode"] == "always_last"
            return
    raise AssertionError("No change dependency found")


def test_tabbed_interface_predictions(connect):
    hello_world = gradio.Interface(
        lambda name: "Hello " + name, "text", "text", api_name="predict"
    )
    bye_world = gradio.Interface(
        lambda name: "Bye " + name, "text", "text", api_name="predict"
    )

    demo = gradio.TabbedInterface(
        [hello_world, bye_world], ["Hello World", "Bye World"]
    )
    with connect(demo) as client:
        assert client.predict("Emily", api_name="/predict") == "Hello Emily"
        assert client.predict("Hannah", api_name="/predict") == "Hello Hannah"


def test_interface_predictions_default_api_name(connect):
    def greet(name):
        return "Hello " + name

    hello_world = gradio.Interface(greet, "text", "text")

    with connect(hello_world) as client:
        assert client.predict("Emily", api_name="/greet") == "Hello Emily"
        assert client.predict("Hannah", api_name="/greet") == "Hello Hannah"

    hello_lambda = gradio.Interface(lambda s: "Hello " + s, "text", "text")

    with connect(hello_lambda) as client:
        assert client.predict("Emily", api_name="/lambda") == "Hello Emily"
        assert client.predict("Hannah", api_name="/lambda") == "Hello Hannah"



"""gr.AnnotatedImage() component."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Literal

import gradio_client.utils as client_utils
import numpy as np
import PIL.Image
from gradio_client import handle_file
from gradio_client.documentation import document

from gradio import processing_utils, utils
from gradio.components.base import Component
from gradio.components.button import Button
from gradio.data_classes import FileData, GradioModel
from gradio.events import Events
from gradio.i18n import I18nData
from gradio.utils import set_default_buttons

if TYPE_CHECKING:
    from gradio.components import Timer

PIL.Image.init()  # fixes https://github.com/gradio-app/gradio/issues/2843


class Annotation(GradioModel):
    image: FileData
    label: str


class AnnotatedImageData(GradioModel):
    image: FileData
    annotations: list[Annotation]

from gradio.events import Dependency

@document()
class AnnotatedImage(Component):
    """
    Creates a component to displays a base image and colored annotations on top of that image. Annotations can take the from of rectangles (e.g. object detection) or masks (e.g. image segmentation).
    As this component does not accept user input, it is rarely used as an input component.

    Demos: image_segmentation
    """

    EVENTS = [Events.select]

    data_model = AnnotatedImageData

    def __init__(
        self,
        value: (
            tuple[
                np.ndarray | PIL.Image.Image | str,
                list[tuple[np.ndarray | tuple[int, int, int, int], str]],
            ]
            | None
        ) = None,
        *,
        format: str = "webp",
        show_legend: bool = True,
        height: int | str | None = None,
        width: int | str | None = None,
        color_map: dict[str, str] | None = None,
        label: str | I18nData | None = None,
        every: Timer | float | None = None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        show_label: bool | None = None,
        container: bool = True,
        scale: int | None = None,
        min_width: int = 160,
        visible: bool | Literal["hidden"] = True,
        elem_id: str | None = None,
        elem_classes: list[str] | str | None = None,
        render: bool = True,
        key: int | str | tuple[int | str, ...] | None = None,
        preserved_by_key: list[str] | str | None = "value",
        buttons: list[Literal["fullscreen"] | Button] | None = None,
    ):
        """
        Parameters:
            value: Tuple of base image and list of (annotation, label) pairs.
            format: Format used to save images before it is returned to the front end, such as 'jpeg' or 'png'. This parameter only takes effect when the base image is returned from the prediction function as a numpy array or a PIL Image. The format should be supported by the PIL library.
            show_legend: If True, will show a legend of the annotations.
            height: The height of the component, specified in pixels if a number is passed, or in CSS units if a string is passed. This has no effect on the preprocessed image file or numpy array, but will affect the displayed image.
            width: The width of the component, specified in pixels if a number is passed, or in CSS units if a string is passed. This has no effect on the preprocessed image file or numpy array, but will affect the displayed image.
            color_map: A dictionary mapping labels to colors. The colors must be specified as hex codes.
            label: the label for this component. Appears above the component and is also used as the header if there are a table of examples for this component. If None and used in a `gr.Interface`, the label will be the name of the parameter this component is assigned to.
            every: Continously calls `value` to recalculate it if `value` is a function (has no effect otherwise). Can provide a Timer whose tick resets `value`, or a float that provides the regular interval for the reset Timer.
            inputs: Components that are used as inputs to calculate `value` if `value` is a function (has no effect otherwise). `value` is recalculated any time the inputs change.
            show_label: if True, will display label.
            container: If True, will place the component in a container - providing some extra padding around the border.
            scale: Relative width compared to adjacent Components in a Row. For example, if Component A has scale=2, and Component B has scale=1, A will be twice as wide as B. Should be an integer.
            min_width: Minimum pixel width, will wrap if not sufficient screen space to satisfy this value. If a certain scale value results in this Component being narrower than min_width, the min_width parameter will be respected first.
            visible: If False, component will be hidden. If "hidden", component will be visually hidden and not take up space in the layout but still exist in the DOM
            elem_id: An optional string that is assigned as the id of this component in the HTML DOM. Can be used for targeting CSS styles.
            elem_classes: An optional list of strings that are assigned as the classes of this component in the HTML DOM. Can be used for targeting CSS styles.
            render: If False, component will not render be rendered in the Blocks context. Should be used if the intention is to assign event listeners now but render the component later.
            key: in a gr.render, Components with the same key across re-renders are treated as the same component, not a new component. Properties set in 'preserved_by_key' are not reset across a re-render.
            preserved_by_key: A list of parameters from this component's constructor. Inside a gr.render() function, if a component is re-rendered with the same key, these (and only these) parameters will be preserved in the UI (if they have been changed by the user or an event listener) instead of re-rendered based on the values provided during constructor.
            buttons: A list of buttons to show for the component. Valid options in the list are "fullscreen" or a gr.Button() instance. The "fullscreen" button allows the user to view the image in fullscreen mode. Custom gr.Button() instances will appear in the toolbar with their configured icon and/or label, and clicking them will trigger any .click() events registered on the button. By default, the fullscreen button is shown, and this can hidden by providing an empty list.
        """
        self.format = format
        self.show_legend = show_legend
        self.height = height
        self.width = width
        self.color_map = color_map
        self.buttons = set_default_buttons(buttons, ["fullscreen"])
        self._value_description = "a tuple of type [image: str, annotations: list[tuple[mask: str, label: str]]] where 'image' is the path to the base image and 'annotations' is a list of tuples where each tuple has a 'mask' image filepath and a corresponding label."
        super().__init__(
            label=label,
            every=every,
            inputs=inputs,
            show_label=show_label,
            container=container,
            scale=scale,
            min_width=min_width,
            visible=visible,
            elem_id=elem_id,
            elem_classes=elem_classes,
            render=render,
            key=key,
            preserved_by_key=preserved_by_key,
            value=value,
        )

    def preprocess(
        self, payload: AnnotatedImageData | None
    ) -> tuple[str, list[tuple[str, str]]] | None:
        """
        Parameters:
            payload: Dict of base image and list of annotations.
        Returns:
            Passes its value as a `tuple` consisting of a `str` filepath to a base image and `list` of annotations. Each annotation itself is `tuple` of a mask (as a `str` filepath to image) and a `str` label.
        """
        if payload is None:
            return None
        base_img = payload.image.path
        annotations = [(a.image.path, a.label) for a in payload.annotations]
        return (base_img, annotations)

    def postprocess(
        self,
        value: (
            tuple[
                np.ndarray | PIL.Image.Image | str,
                Sequence[tuple[np.ndarray | tuple[int, int, int, int], str]],
            ]
            | None
        ),
    ) -> AnnotatedImageData | None:
        """
        Parameters:
            value: Expects a a tuple of a base image and list of annotations: a `tuple[Image, list[Annotation]]`. The `Image` itself can be `str` filepath, `numpy.ndarray`, or `PIL.Image`. Each `Annotation` is a `tuple[Mask, str]`. The `Mask` can be either a `tuple` of 4 `int`'s representing the bounding box coordinates (x1, y1, x2, y2), or 0-1 confidence mask in the form of a `numpy.ndarray` of the same shape as the image, while the second element of the `Annotation` tuple is a `str` label.
        Returns:
            Tuple of base image file and list of annotations, with each annotation a two-part tuple where the first element image path of the mask, and the second element is the label.
        """
        if value is None:
            return None
        base_img = value[0]
        if isinstance(base_img, str):
            if client_utils.is_http_url_like(base_img):
                base_img = processing_utils.save_url_to_cache(
                    base_img, cache_dir=self.GRADIO_CACHE
                )
            base_img_path = base_img
            base_img = np.array(PIL.Image.open(base_img))
        elif isinstance(base_img, np.ndarray):
            base_file = processing_utils.save_img_array_to_cache(
                base_img, cache_dir=self.GRADIO_CACHE, format=self.format
            )
            base_img_path = str(utils.abspath(base_file))
        elif isinstance(base_img, PIL.Image.Image):
            base_file = processing_utils.save_pil_to_cache(
                base_img, cache_dir=self.GRADIO_CACHE, format=self.format
            )
            base_img_path = str(utils.abspath(base_file))
            base_img = np.array(base_img)
        else:
            raise ValueError(
                "AnnotatedImage only accepts filepaths, PIL images or numpy arrays for the base image."
            )

        sections = []
        color_map = self.color_map or {}

        def hex_to_rgb(value):
            value = value.lstrip("#")
            lv = len(value)
            return [int(value[i : i + lv // 3], 16) for i in range(0, lv, lv // 3)]

        for mask, label in value[1]:
            mask_array = np.zeros((base_img.shape[0], base_img.shape[1]))
            if isinstance(mask, np.ndarray):
                mask_array = mask
            else:
                x1, y1, x2, y2 = mask
                border_width = 3
                mask_array[y1:y2, x1:x2] = 0.5
                mask_array[y1:y2, x1 : x1 + border_width] = 1
                mask_array[y1:y2, x2 - border_width : x2] = 1
                mask_array[y1 : y1 + border_width, x1:x2] = 1
                mask_array[y2 - border_width : y2, x1:x2] = 1

            if label in color_map:
                rgb_color = hex_to_rgb(color_map[label])
            else:
                rgb_color = [255, 0, 0]
            colored_mask = np.zeros((base_img.shape[0], base_img.shape[1], 4))
            solid_mask = np.copy(mask_array)
            solid_mask[solid_mask > 0] = 1

            colored_mask[:, :, 0] = rgb_color[0] * solid_mask
            colored_mask[:, :, 1] = rgb_color[1] * solid_mask
            colored_mask[:, :, 2] = rgb_color[2] * solid_mask
            colored_mask[:, :, 3] = mask_array * 255

            colored_mask_img = PIL.Image.fromarray((colored_mask).astype(np.uint8))

            # RGBA does not support transparency
            mask_file = processing_utils.save_pil_to_cache(
                colored_mask_img, cache_dir=self.GRADIO_CACHE, format="png"
            )
            mask_file_path = str(utils.abspath(mask_file))
            sections.append(
                Annotation(image=FileData(path=mask_file_path), label=label)
            )

        return AnnotatedImageData(
            image=FileData(path=base_img_path),
            annotations=sections,
        )

    def example_payload(self) -> Any:
        return {
            "image": handle_file(
                "https://raw.githubusercontent.com/gradio-app/gradio/main/test/test_files/bus.png"
            ),
            "annotations": [],
        }

    def example_value(self) -> Any:
        return (
            "https://raw.githubusercontent.com/gradio-app/gradio/main/test/test_files/bus.png",
            [([0, 0, 100, 100], "bus")],
        )
    from typing import Callable, Literal, Sequence, Any, TYPE_CHECKING
    from gradio.blocks import Block
    if TYPE_CHECKING:
        from gradio.components import Timer
        from gradio.components.base import Component

    
    def select(self,
        fn: Callable[..., Any] | None = None,
        inputs: Block | Sequence[Block] | set[Block] | None = None,
        outputs: Block | Sequence[Block] | None = None,
        api_name: str | None = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        show_progress_on: Component | Sequence[Component] | None = None,
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: Timer | float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | Literal[True] | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        api_visibility: Literal["public", "private", "undocumented"] = "public",
        key: int | str | tuple[int | str, ...] | None = None,
        api_description: str | None | Literal[False] = None,
        validator: Callable[..., Any] | None = None,
    
        ) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: list of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: list of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: defines how the endpoint appears in the API docs. Can be a string or None. If set to a string, the endpoint will be exposed in the API docs with the given name. If None (default), the name of the function will be used as the API endpoint.
            scroll_to_output: if True, will scroll to output component on completion
            show_progress: how to show the progress animation while event is running: "full" shows a spinner which covers the output component area as well as a runtime display in the upper right corner, "minimal" only shows the runtime display, "hidden" shows no progress animation at all
            show_progress_on: Component or list of components to show the progress animation on. If None, will show the progress animation on all of the output components.
            queue: if True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: if True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: if False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: if False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: a list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: continously calls `value` to recalculate it if `value` is a function (has no effect otherwise). Can provide a Timer whose tick resets `value`, or a float that provides the regular interval for the reset Timer.
            trigger_mode: if "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` and `.key_up()` events) would allow a second submission after the pending event is complete.
            js: optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: if set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: if set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            api_visibility: controls the visibility and accessibility of this endpoint. Can be "public" (shown in API docs and callable by clients), "private" (hidden from API docs and not callable by clients), or "undocumented" (hidden from API docs but callable by clients and via gr.load). If fn is None, api_visibility will automatically be set to "private".
            key: A unique key for this event listener to be used in @gr.render(). If set, this value identifies an event as identical across re-renders when the key is identical.
            api_description: Description of the API endpoint. Can be a string, None, or False. If set to a string, the endpoint will be exposed in the API docs with the given description. If None, the function's docstring will be used as the API endpoint description. If False, then no description will be displayed in the API docs.
            validator: Optional validation function to run before the main function. If provided, this function will be executed first with queue=False, and only if it completes successfully will the main function be called. The validator receives the same inputs as the main function.
        
        """
        ...


import os

from huggingface_hub import interpreter_login, whoami


def save_login():
    try:
        while True:
            response = input("Stay logged in to Hugging Face? (Y/n): ").strip().lower()
            if response in ["y", "yes", ""]:  # Empty string for just pressing Enter
                os.environ["GRADIO_AUTO_LOGOUT"] = "false"
                print("You can logout at any time with: hf auth logout")
                break
            elif response in ["n", "no"]:
                os.environ["GRADIO_AUTO_LOGOUT"] = "true"
                break
            else:
                print("Please answer with 'y' or 'n'")
    except (EOFError, KeyboardInterrupt):
        os.environ["GRADIO_AUTO_LOGOUT"] = "true"


def hf_login():
    if os.getenv("HF_TOKEN"):
        try:
            user = whoami(token=os.getenv("HF_TOKEN")).get("name")
            print(
                f"🔓  Logged in to Hugging Face as {user}. You can logout at any time with: unset HF_TOKEN\n"
            )
        except Exception:
            print("❌  Error logging in to Hugging Face with $HF_TOKEN")
            print("Logging in with CLI prompt...\n")
            interpreter_login()
            save_login()
    else:
        try:
            user = whoami().get("name")
            print(
                f"🔓  Logged in to Hugging Face as {user}. You can logout at any time with: hf auth logout\n"
            )
        except Exception:
            print(
                "🔑  No Hugging Face login found, launching login prompt... \n\nPlease use a token with permission to make calls to Inference Providers \n"
            )
            interpreter_login()
            save_login()



from .accordion import Accordion
from .column import Column
from .draggable import Draggable
from .form import Form
from .group import Group
from .row import Row
from .sidebar import Sidebar
from .tabs import Tab, TabItem, Tabs
from .walkthrough import Step, Walkthrough

__all__ = [
    "Accordion",
    "Column",
    "Draggable",
    "Form",
    "Row",
    "Group",
    "Tabs",
    "Tab",
    "TabItem",
    "Sidebar",
    "Walkthrough",
    "Step",
]



from __future__ import annotations

import json
from collections.abc import Callable, Sequence, Set
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
)

from gradio_client.documentation import document

from gradio.components.base import Component
from gradio.components.button import Button
from gradio.data_classes import GradioModel
from gradio.events import Events
from gradio.i18n import I18nData
from gradio.utils import set_default_buttons

if TYPE_CHECKING:
    import pandas as pd

    from gradio.components import Timer


class PlotData(GradioModel):
    columns: list[str]
    data: list[list[Any]]
    datatypes: dict[str, Literal["quantitative", "nominal", "temporal"]]
    mark: str


class NativePlot(Component):
    """
    Creates a native Gradio plot component to display data from a pandas DataFrame. Supports interactivity and updates.

    Demos: native_plots
    """

    EVENTS = [Events.select, Events.double_click]

    def __init__(
        self,
        value: pd.DataFrame | Callable | None = None,
        x: str | None = None,
        y: str | None = None,
        *,
        color: str | None = None,
        title: str | None = None,
        x_title: str | None = None,
        y_title: str | None = None,
        color_title: str | None = None,
        x_bin: str | float | None = None,
        y_aggregate: Literal["sum", "mean", "median", "min", "max", "count"]
        | None = None,
        color_map: dict[str, str] | None = None,
        colors_in_legend: list[str] | None = None,
        x_lim: list[float | None] | None = None,
        y_lim: list[float | None] = None,
        x_label_angle: float = 0,
        y_label_angle: float = 0,
        x_axis_labels_visible: bool | Literal["hidden"] = True,
        caption: str | I18nData | None = None,
        sort: Literal["x", "y", "-x", "-y"] | list[str] | None = None,
        tooltip: Literal["axis", "none", "all"] | list[str] = "axis",
        height: int | None = None,
        label: str | I18nData | None = None,
        show_label: bool | None = None,
        container: bool = True,
        scale: int | None = None,
        min_width: int = 160,
        every: Timer | float | None = None,
        inputs: Component | Sequence[Component] | Set[Component] | None = None,
        visible: bool | Literal["hidden"] = True,
        elem_id: str | None = None,
        elem_classes: list[str] | str | None = None,
        render: bool = True,
        buttons: list[Literal["fullscreen", "export"] | Button] | None = None,
        key: int | str | tuple[int | str, ...] | None = None,
        preserved_by_key: list[str] | str | None = "value",
    ):
        """
        Parameters:
            value: The pandas dataframe containing the data to display in the plot.
            x: Column corresponding to the x axis. Column can be numeric, datetime, or string/category.
            y: Column corresponding to the y axis. Column must be numeric.
            color: Column corresponding to series, visualized by color. Column must be string/category.
            title: The title to display on top of the chart.
            x_title: The title given to the x axis. By default, uses the value of the x parameter.
            y_title: The title given to the y axis. By default, uses the value of the y parameter.
            color_title: The title given to the color legend. By default, uses the value of color parameter.
            x_bin: Grouping used to cluster x values. If x column is numeric, should be number to bin the x values. If x column is datetime, should be string such as "1h", "15m", "10s", using "s", "m", "h", "d" suffixes.
            y_aggregate: Aggregation function used to aggregate y values, used if x_bin is provided or x is a string/category. Must be one of "sum", "mean", "median", "min", "max".
            color_map: Mapping of series to color names or codes. For example, {"success": "green", "fail": "#FF8888"}.
            colors_in_legend: List containing column names of the series to show in the legend. By default, all series are shown.
            height: The height of the plot in pixels.
            x_lim: A tuple or list containing the limits for the x-axis, specified as [x_min, x_max]. To fix only one of these values, set the other to None, e.g. [0, None] to scale from 0 to the maximum value. If x column is datetime type, x_lim should be timestamps.
            y_lim: A tuple of list containing the limits for the y-axis, specified as [y_min, y_max]. To fix only one of these values, set the other to None, e.g. [0, None] to scale from 0 to the maximum to value.
            x_label_angle: The angle of the x-axis labels in degrees offset clockwise.
            y_label_angle: The angle of the y-axis labels in degrees offset clockwise.
            x_axis_labels_visible: Whether the x-axis labels should be visible. Can be hidden when many x-axis labels are present.
            caption: The (optional) caption to display below the plot.
            sort: The sorting order of the x values, if x column is type string/category. Can be "x", "y", "-x", "-y", or list of strings that represent the order of the categories.
            tooltip: The tooltip to display when hovering on a point. "axis" shows the values for the axis columns, "all" shows all column values, and "none" shows no tooltips. Can also provide a list of strings representing columns to show in the tooltip, which will be displayed along with axis values.
            height: The height of the plot in pixels.
            label: The (optional) label to display on the top left corner of the plot.
            show_label: Whether the label should be displayed.
            container: If True, will place the component in a container - providing some extra padding around the border.
            scale: relative size compared to adjacent Components. For example if Components A and B are in a Row, and A has scale=2, and B has scale=1, A will be twice as wide as B. Should be an integer. scale applies in Rows, and to top-level Components in Blocks where fill_height=True.
            min_width: minimum pixel width, will wrap if not sufficient screen space to satisfy this value. If a certain scale value results in this Component being narrower than min_width, the min_width parameter will be respected first.
            every: Continously calls `value` to recalculate it if `value` is a function (has no effect otherwise). Can provide a Timer whose tick resets `value`, or a float that provides the regular interval for the reset Timer.
            inputs: Components that are used as inputs to calculate `value` if `value` is a function (has no effect otherwise). `value` is recalculated any time the inputs change.
            visible: Whether the plot should be visible.
            elem_id: An optional string that is assigned as the id of this component in the HTML DOM. Can be used for targeting CSS styles.
            elem_classes: An optional list of strings that are assigned as the classes of this component in the HTML DOM. Can be used for targeting CSS styles.
            render: If False, component will not render be rendered in the Blocks context. Should be used if the intention is to assign event listeners now but render the component later.
            buttons: A list of buttons to show for the component. Valid options are "fullscreen", "export", or a gr.Button() instance. The "fullscreen" button allows the user to view the plot in fullscreen mode. The "export" button allows the user to export and download the current view of the plot as a PNG image. Custom gr.Button() instances will appear in the toolbar with their configured icon and/or label, and clicking them will trigger any .click() events registered on the button. By default, no buttons are shown.
            key: in a gr.render, Components with the same key across re-renders are treated as the same component, not a new component. Properties set in 'preserved_by_key' are not reset across a re-render.
            preserved_by_key: A list of parameters from this component's constructor. Inside a gr.render() function, if a component is re-rendered with the same key, these (and only these) parameters will be preserved in the UI (if they have been changed by the user or an event listener) instead of re-rendered based on the values provided during constructor.
        """
        self.x = x
        self.y = y
        self.color = color
        self.title = title
        self.x_title = x_title
        self.y_title = y_title
        self.color_title = color_title
        self.x_bin = x_bin
        self.y_aggregate = y_aggregate
        self.color_map = color_map
        self.x_lim = x_lim
        self.y_lim = y_lim
        self.x_label_angle = x_label_angle
        self.y_label_angle = y_label_angle
        self.x_axis_labels_visible = x_axis_labels_visible
        self.caption = caption
        self.sort = sort
        self.tooltip = tooltip
        self.height = height
        self.buttons = set_default_buttons(buttons, None)
        self.colors_in_legend = colors_in_legend

        if label is None and show_label is None:
            show_label = False
        super().__init__(
            value=value,
            label=label,
            show_label=show_label,
            container=container,
            scale=scale,
            min_width=min_width,
            visible=visible,
            elem_id=elem_id,
            elem_classes=elem_classes,
            render=render,
            key=key,
            preserved_by_key=preserved_by_key,
            every=every,
            inputs=inputs,
        )

    def get_block_name(self) -> str:
        return "nativeplot"

    def get_mark(self) -> str:
        return "native"

    def preprocess(self, payload: PlotData | None) -> PlotData | None:
        """
        Parameters:
            payload: The data to display in a line plot.
        Returns:
            The data to display in a line plot.
        """
        return payload

    def postprocess(self, value: pd.DataFrame | dict | None) -> PlotData | None:
        """
        Parameters:
            value: Expects a pandas DataFrame containing the data to display in the line plot. The DataFrame should contain at least two columns, one for the x-axis (corresponding to this component's `x` argument) and one for the y-axis (corresponding to `y`).
        Returns:
            The data to display in a line plot, in the form of an AltairPlotData dataclass, which includes the plot information as a JSON string, as well as the type of plot (in this case, "line").
        """
        # if None or update
        if value is None or isinstance(value, dict):
            return value

        def get_simplified_type(dtype):
            import pandas as pd

            if pd.api.types.is_numeric_dtype(dtype):
                return "quantitative"
            elif pd.api.types.is_string_dtype(
                dtype
            ) or pd.api.types.is_categorical_dtype(dtype):
                return "nominal"
            elif pd.api.types.is_datetime64_any_dtype(dtype):
                return "temporal"
            else:
                raise ValueError(f"Unsupported data type: {dtype}")

        split_json = json.loads(value.to_json(orient="split", date_unit="ms"))
        datatypes = {
            col: get_simplified_type(value[col].dtype) for col in value.columns
        }
        return PlotData(
            columns=split_json["columns"],
            data=split_json["data"],
            datatypes=datatypes,
            mark=self.get_mark(),
        )

    def example_payload(self) -> Any:
        return None

    def example_value(self) -> Any:
        import pandas as pd

        return pd.DataFrame({self.x: [1, 2, 3], self.y: [4, 5, 6]})

    def api_info(self) -> dict[str, Any]:
        return {"type": {}, "description": "any valid json"}


@document()
class BarPlot(NativePlot):
    """
    Creates a bar plot component to display data from a pandas DataFrame.

    Demos: bar_plot_demo
    """

    def get_block_name(self) -> str:
        return "nativeplot"

    def get_mark(self) -> str:
        return "bar"


@document()
class LinePlot(NativePlot):
    """
    Creates a line plot component to display data from a pandas DataFrame.

    Demos: line_plot_demo
    """

    def get_block_name(self) -> str:
        return "nativeplot"

    def get_mark(self) -> str:
        return "line"


@document()
class ScatterPlot(NativePlot):
    """
    Creates a scatter plot component to display data from a pandas DataFrame.

    Demos: scatter_plot_demo
    """

    def get_block_name(self) -> str:
        return "nativeplot"

    def get_mark(self) -> str:
        return "point"



import base64
import contextlib
import copy
import os
import re
import tempfile
import warnings
from collections.abc import AsyncIterator, Sequence
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, cast
from urllib.parse import unquote

import gradio_client.utils as client_utils
import httpx
from anyio.to_thread import run_sync
from gradio_client import Client, handle_file
from gradio_client.utils import Status, StatusUpdate
from PIL import Image
from pydantic import AnyUrl
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.routing import Mount, Route
from starlette.types import Receive, Scope, Send

from gradio import processing_utils, route_utils, utils
from gradio.blocks import BlockFunction
from gradio.components import State
from gradio.route_utils import Header

if TYPE_CHECKING:
    from mcp import types  # noqa: F401
    from mcp.server import Server  # noqa: F401
    from mcp.server.lowlevel.helper_types import ReadResourceContents  # noqa: F401

    from gradio.blocks import BlockContext, Blocks
    from gradio.components import Component


DEFAULT_TEMP_DIR = os.environ.get("GRADIO_TEMP_DIR") or str(
    Path(tempfile.gettempdir()) / "gradio"
)


class GradioMCPServer:
    """
    A class for creating an MCP server around a Gradio app. This class
    requires `mcp` to be installed.

    Args:
        blocks: The Blocks app to create the MCP server for.
    """

    # Imports are here to avoid needing to install `mcp` when not using this class.
    # This way, we are able to export `gr.tool`, `gr.resource`, etc. to `__init__.py`
    # without the user needing to have `mcp` installed.
    try:
        from mcp import types
        from mcp.server import Server
        from mcp.server.lowlevel.helper_types import ReadResourceContents
        from mcp.server.sse import SseServerTransport
        from mcp.server.streamable_http_manager import StreamableHTTPSessionManager
    except ImportError:
        pass

    def __init__(self, blocks: "Blocks"):
        try:
            import mcp  # noqa: F401
        except ImportError as e:
            raise ImportError(
                "The `mcp` package is required to use the Gradio MCP integration. Please install it with the `mcp` extra: `pip install gradio[mcp]`."
            ) from e

        self.blocks = blocks
        self.api_info = self.blocks.get_api_info()
        self.mcp_server = self.create_mcp_server()
        self.root_path = ""
        space_id = utils.get_space()
        self.tool_prefix = space_id.split("/")[-1] + "_" if space_id else ""
        self.tool_to_endpoint = self.get_tool_to_endpoint()
        self.warn_about_state_inputs()
        self._local_url: str | None = None
        self._client_instance: Client | None = None

        manager = self.StreamableHTTPSessionManager(  # type: ignore
            app=self.mcp_server, json_response=False, stateless=True
        )

        async def handle_streamable_http(
            scope: Scope, receive: Receive, send: Send
        ) -> None:
            path = scope.get("path", "")
            if not path.endswith(
                (
                    "/gradio_api/mcp",
                    "/gradio_api/mcp/",
                    "/gradio_api/mcp/http",
                    "/gradio_api/mcp/http/",
                )
            ):
                response = Response(
                    content=f"Path '{path}' not found. The MCP HTTP transport is available at /gradio_api/mcp.",
                    status_code=404,
                )
                await response(scope, receive, send)
                return

            await manager.handle_request(scope, receive, send)

        @contextlib.asynccontextmanager
        async def lifespan(app: Starlette) -> AsyncIterator[None]:  # noqa: ARG001
            """Context manager for managing session manager lifecycle."""
            async with manager.run():
                try:
                    yield
                finally:
                    pass

        self.lifespan = lifespan
        self.manager = manager
        self.handle_streamable_http = handle_streamable_http

    @property
    def local_url(self) -> str | None:
        return self._local_url

    def get_route_path(self, request: Request) -> str:  # type: ignore
        """
        Gets the route path of the MCP server based on the incoming request.
        Can be different depending on whether the request is coming from the MCP SSE transport or the HTTP transport.
        """
        url = httpx.URL(str(request.url))
        url = url.copy_with(query=None)
        url = str(url).rstrip("/")
        if url.endswith("/gradio_api/mcp/messages"):
            return "/gradio_api/mcp/messages"
        else:
            return "/gradio_api/mcp"

    def get_selected_tools_from_request(self) -> list[str] | None:
        """
        Extract the selected tools from the request query parameters and return the full tool names (with the tool prefix).
        Returns None if no tools parameter is specified (meaning all tools are available).
        """
        context_request: Request | None = self.mcp_server.request_context.request
        if context_request is None:
            return None
        query_params = dict(getattr(context_request, "query_params", {}))
        if "tools" in query_params:
            tools = query_params["tools"].split(",")
            full_tool_names = [self.tool_prefix + tool for tool in tools]
            return full_tool_names
        return None

    @staticmethod
    def valid_and_unique_tool_name(
        tool_name: str, existing_tool_names: set[str]
    ) -> str:
        """
        Sanitizes a tool name to make it a valid MCP tool name (only
        alphanumeric characters, underscores, <= 128 characters)
        and is unique among the existing tool names.
        """
        tool_name = re.sub(r"[^a-zA-Z0-9]", "_", tool_name)
        tool_name = tool_name[:120]  # Leave room for suffix if needed
        tool_name_base = tool_name
        suffix = 1
        while tool_name in existing_tool_names:
            tool_name = tool_name_base + f"_{suffix}"
            suffix += 1
        return tool_name

    def get_tool_to_endpoint(self) -> dict[str, str]:
        """
        Gets all of the tools that are exposed by the Gradio app and also
        creates a mapping from the tool names to the endpoint names in the API docs.
        """
        tool_to_endpoint = {}
        for endpoint_name in self.api_info["named_endpoints"]:
            block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
            if block_fn is None or block_fn.fn is None:
                continue
            fn_name = (
                getattr(block_fn.fn, "__name__", None)
                or (
                    hasattr(block_fn.fn, "__class__")
                    and getattr(block_fn.fn.__class__, "__name__", None)
                )
                or endpoint_name.lstrip("/")
            )
            tool_name = self.tool_prefix + fn_name
            tool_name = self.valid_and_unique_tool_name(
                tool_name, set(tool_to_endpoint.keys())
            )
            tool_to_endpoint[tool_name] = endpoint_name
        return tool_to_endpoint

    def warn_about_state_inputs(self) -> None:
        """
        Warn about tools that have gr.State inputs.
        """
        for _, endpoint_name in self.tool_to_endpoint.items():
            block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
            if block_fn and any(isinstance(input, State) for input in block_fn.inputs):
                warnings.warn(
                    "This MCP server includes a tool that has a gr.State input, which will not be "
                    "updated between tool calls. The original, default value of the State will be "
                    "used each time."
                )

    def _get_or_create_client(self) -> Client:
        if self._client_instance is None:
            context_request: Request | None = self.mcp_server.request_context.request
            if context_request is None:
                raise ValueError(
                    "Could not find the request object in the MCP server context. This is not expected to happen. Please raise an issue: https://github.com/gradio-app/gradio."
                )
            route_path = self.get_route_path(context_request)
            root_url = route_utils.get_root_url(
                request=context_request,
                route_path=route_path,
                root_path=self.root_path,
            )
            self._client_instance = Client(
                self.local_url or root_url,
                download_files=False,
                verbose=False,
                analytics_enabled=False,
                ssl_verify=False,
                _skip_components=False,
                headers={"x-gradio-user": "mcp"},
            )
        return self._client_instance

    def _prepare_tool_call_args(
        self, name: str, arguments: dict[str, Any]
    ) -> tuple[str, list[Any], dict[str, str], "BlockFunction"]:
        """
        Prepare and validate arguments for a tool call.

        Returns:
            A tuple of (endpoint_name, processed_args, request_headers, block_fn)
        """
        selected_tools = self.get_selected_tools_from_request()
        _, filedata_positions = self.get_input_schema(name)
        processed_kwargs = self.convert_strings_to_filedata(
            arguments, filedata_positions
        )
        endpoint_name = self.tool_to_endpoint.get(name)
        if endpoint_name is None:
            raise ValueError(f"Unknown tool for this Gradio app: {name}")

        if selected_tools is not None and name not in selected_tools:
            raise ValueError(f"Tool '{name}' is not in the selected tools list")

        block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
        assert block_fn is not None  # noqa: S101

        if endpoint_name in self.api_info["named_endpoints"]:
            parameters_info = self.api_info["named_endpoints"][endpoint_name][
                "parameters"
            ]
            processed_args = client_utils.construct_args(
                parameters_info,
                (),
                processed_kwargs,
            )
        else:
            processed_args = []

        context_request: Request | None = self.mcp_server.request_context.request
        if context_request is None:
            raise ValueError(
                "Could not find the request object in the MCP server context. This is not expected to happen. Please raise an issue: https://github.com/gradio-app/gradio."
            )
        request_headers = dict(context_request.headers.items())
        request_headers.pop("content-length", None)
        request_headers.pop("x-gradio-user", None)

        return endpoint_name, processed_args, request_headers, block_fn

    async def _execute_tool_without_progress(self, job: Any) -> list[Any]:
        """
        Execute a tool call without progress tracking (fast path).

        Calls job.result() to get the final output without processing
        intermediate status updates.

        Returns:
            The output data as a list.
        """
        result = await run_sync(job.result)
        return [result]

    @staticmethod
    def _format_progress_message(update: StatusUpdate) -> str | None:
        """
        Format a status update into a human-readable progress message.

        Returns:
            A formatted message string, or None if no message should be shown.
        """
        if update.code in [Status.JOINING_QUEUE, Status.STARTING]:
            return "Joined server queue."
        elif update.code in [Status.IN_QUEUE]:
            message = f"In queue. Position {update.rank} out of {update.queue_size}."
            if update.eta is not None:
                message += f" Estimated time remaining: {update.eta} seconds."
            return message
        elif update.code in [Status.PROGRESS]:
            for progress_unit in update.progress_data or []:
                title = (
                    "Progress"
                    if progress_unit.desc is None
                    else f"Progress {progress_unit.desc}"
                )
                if progress_unit.index is not None and progress_unit.length is not None:
                    return (
                        f"{title}: Step {progress_unit.index} of {progress_unit.length}"
                    )
                elif progress_unit.index is not None and progress_unit.length is None:
                    return f"{title}: Step {progress_unit.index}"
        elif update.code in [Status.PROCESSING, Status.ITERATING]:
            return "Processing"
        return None

    async def _execute_tool_with_progress(  # type: ignore
        self, job: Any, progress_token: str
    ) -> dict[str, Any]:
        """
        Execute a tool call with progress tracking (streaming path).

        Iterates through job updates to send progress notifications to the client.

        Returns:
            The output data as a list.
        """
        step = 0
        async for update in job:
            if update.type == "status":
                update = cast(StatusUpdate, update)
                message = self._format_progress_message(update)

                await (
                    self.mcp_server.request_context.session.send_progress_notification(
                        progress_token=progress_token,
                        progress=step,
                        message=message,  # type: ignore
                        related_request_id=str(
                            self.mcp_server.request_context.request_id
                        ),
                    )
                )
                step += 1
            elif update.type == "output" and update.final:
                output = update.outputs
                if not update.success:
                    error_title = output.get("title")
                    error_message = output.get("error")
                    if error_title and error_message:
                        msg = f"{error_title}: {error_message}"
                    elif error_message:
                        msg = error_message
                    elif error_title:
                        msg = error_title
                    else:
                        msg = "Error!"
                    raise RuntimeError(msg)
        if job.exception():
            raise job.exception()
        return output["data"]

    def create_mcp_server(self) -> "Server":
        """
        Create an MCP server for the given Gradio Blocks app.

        Parameters:
            blocks: The Blocks app to create the MCP server for.

        Returns:
            The MCP server.
        """
        server = self.Server(str(self.blocks.title or "Gradio App"))  # type: ignore

        @server.call_tool()
        async def call_tool(
            name: str, arguments: dict[str, Any]
        ) -> self.types.CallToolResult:  # type: ignore
            """
            Call a tool on the Gradio app.

            Args:
                name: The name of the tool to call.
                arguments: The arguments to pass to the tool.
            """
            progress_token = None
            if self.mcp_server.request_context.meta is not None:
                progress_token = self.mcp_server.request_context.meta.progressToken

            client = await run_sync(self._get_or_create_client)
            endpoint_name, processed_args, request_headers, block_fn = (
                self._prepare_tool_call_args(name, arguments)
            )
            processed_args = self.insert_empty_state(block_fn.inputs, processed_args)
            job = client.submit(
                *processed_args, api_name=endpoint_name, headers=request_headers
            )

            if progress_token is None or not block_fn.queue:
                output_data = await self._execute_tool_without_progress(job)
            else:
                output_data = await self._execute_tool_with_progress(  # type: ignore
                    job,
                    progress_token,  # type: ignore
                )

            output_data = self.pop_returned_state(block_fn.outputs, output_data)

            context_request: Request | None = self.mcp_server.request_context.request
            route_path = self.get_route_path(context_request)  # type: ignore
            root_url = route_utils.get_root_url(  # type: ignore
                request=context_request,  # type: ignore
                route_path=route_path,  # type: ignore
                root_path=self.root_path,  # type: ignore
            )
            content = self.postprocess_output_data(output_data, root_url)
            if getattr(block_fn.fn, "_mcp_structured_output", False):
                structured_content = {"result": content}
            else:
                structured_content = None
            return self.types.CallToolResult(  # type: ignore
                content=content,  # type: ignore
                structuredContent=structured_content,  # type: ignore
                _meta=getattr(block_fn.fn, "_mcp_meta", None),  # type: ignore
            )

        @server.list_tools()
        async def list_tools() -> list[self.types.Tool]:  # type: ignore
            """
            List all tools on the Gradio app.
            """
            selected_tools = self.get_selected_tools_from_request()

            tools = []
            for tool_name, endpoint_name in self.tool_to_endpoint.items():
                if selected_tools is not None and tool_name not in selected_tools:
                    continue

                block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
                if (
                    block_fn is None
                    or block_fn.fn is None
                    or (
                        hasattr(block_fn.fn, "_mcp_type")
                        and block_fn.fn._mcp_type != "tool"
                    )
                ):
                    continue

                description, parameters = self.get_fn_description(block_fn, tool_name)
                schema, _ = self.get_input_schema(tool_name, parameters)
                tool_meta = getattr(block_fn.fn, "_mcp_meta", None)

                tools.append(
                    self.types.Tool(  # type: ignore
                        name=tool_name,
                        description=description,
                        inputSchema=schema,
                        _meta=tool_meta,  # type: ignore
                    )
                )
            return tools

        @server.list_resources()
        async def list_resources() -> list[self.types.Resource]:  # type: ignore
            """
            List all available resources.
            """
            resources = []

            selected_tools = self.get_selected_tools_from_request()
            for tool_name, endpoint_name in self.tool_to_endpoint.items():
                if selected_tools is not None and tool_name not in selected_tools:
                    continue

                block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
                if (
                    block_fn
                    and block_fn.fn
                    and hasattr(block_fn.fn, "_mcp_type")
                    and block_fn.fn._mcp_type == "resource"
                ):
                    uri_template = block_fn.fn._mcp_uri_template  # type: ignore
                    parameters = re.findall(r"\{([^}]+)\}", uri_template)
                    description, parameters, _ = utils.get_function_description(
                        block_fn.fn
                    )
                    if not parameters:
                        resources.append(
                            self.types.Resource(  # type: ignore
                                uri=uri_template,
                                name=block_fn.fn.__name__,  # type: ignore
                                description=description,
                                mimeType=block_fn.fn._mcp_mime_type,  # type: ignore
                            )
                        )
            return resources

        @server.list_resource_templates()
        async def list_resource_templates() -> list[self.types.ResourceTemplate]:  # type: ignore
            """
            List all available resource templates.
            """
            templates = []
            selected_tools = self.get_selected_tools_from_request()
            for tool_name, endpoint_name in self.tool_to_endpoint.items():
                if selected_tools is not None and tool_name not in selected_tools:
                    continue

                block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
                if (
                    block_fn
                    and block_fn.fn
                    and hasattr(block_fn.fn, "_mcp_type")
                    and block_fn.fn._mcp_type == "resource"
                ):
                    uri_template = block_fn.fn._mcp_uri_template  # type: ignore
                    parameters = re.findall(r"\{([^}]+)\}", uri_template)
                    description, parameters, _ = utils.get_function_description(
                        block_fn.fn
                    )
                    if parameters:
                        templates.append(
                            self.types.ResourceTemplate(  # type: ignore
                                uriTemplate=uri_template,
                                name=block_fn.fn.__name__,  # type: ignore
                                description=description,
                                mimeType=block_fn.fn._mcp_mime_type,  # type: ignore
                            )
                        )
            return templates

        @server.read_resource()
        async def read_resource(uri: AnyUrl | str) -> list[self.ReadResourceContents]:  # type: ignore
            """
            Read a specific resource by URI.
            """
            uri = str(uri)
            client = await run_sync(self._get_or_create_client)
            for endpoint_name in self.tool_to_endpoint.values():
                block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)

                if (
                    block_fn
                    and block_fn.fn
                    and hasattr(block_fn.fn, "_mcp_type")
                    and block_fn.fn._mcp_type == "resource"
                ):
                    uri_template = block_fn.fn._mcp_uri_template  # type: ignore
                    parameters = re.findall(r"\{([^}]+)\}", uri_template)  # type: ignore

                    kwargs = {}
                    matched = False

                    if parameters:
                        pattern = re.escape(uri_template)
                        for param in parameters:
                            pattern = pattern.replace(
                                f"\\{{{param}\\}}", f"(?P<{param}>[^/]+)"
                            )
                        match = re.match(f"^{pattern}$", uri)
                        if match:
                            kwargs = match.groupdict()
                            matched = True
                    elif uri_template == uri:
                        matched = True

                    if matched:
                        if endpoint_name in self.api_info["named_endpoints"]:
                            parameters_info = self.api_info["named_endpoints"][
                                endpoint_name
                            ]["parameters"]
                            processed_args = client_utils.construct_args(
                                parameters_info,
                                (),
                                kwargs,
                            )
                        else:
                            processed_args = list(kwargs.values())

                        async for update in client.submit(
                            *processed_args, api_name=endpoint_name
                        ):
                            if update.type == "output" and update.final:  # type: ignore
                                output = update.outputs  # type: ignore
                                result = output["data"][0]
                                break

                        mime_type = block_fn.fn._mcp_mime_type  # type: ignore
                        if mime_type and not mime_type.startswith("text/"):
                            result = base64.b64decode(result.encode("ascii"))
                        return [
                            self.ReadResourceContents(  # type: ignore
                                content=result, mime_type=mime_type
                            )
                        ]

            raise ValueError(f"Resource not found: {uri}")

        @server.list_prompts()
        async def list_prompts() -> list[self.types.Prompt]:  # type: ignore
            """
            List all available prompts.
            """
            prompts = []
            selected_tools = self.get_selected_tools_from_request()
            for tool_name, endpoint_name in self.tool_to_endpoint.items():
                if selected_tools is not None and tool_name not in selected_tools:
                    continue

                block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
                if (
                    block_fn
                    and block_fn.fn
                    and hasattr(block_fn.fn, "_mcp_type")
                    and block_fn.fn._mcp_type == "prompt"
                ):
                    description, parameters, _ = utils.get_function_description(
                        block_fn.fn
                    )
                    function_params = utils.get_function_params(block_fn.fn)
                    arguments = [
                        self.types.PromptArgument(  # type: ignore
                            name=param_name,
                            description=parameters.get(param_name, ""),
                            required=not has_default,
                        )
                        for param_name, has_default, _, _ in function_params
                    ]
                    prompts.append(
                        self.types.Prompt(  # type: ignore
                            name=tool_name,
                            description=description,
                            arguments=arguments,
                        )
                    )
            return prompts

        @server.get_prompt()
        async def get_prompt(
            name: str, arguments: dict[str, Any] | None = None
        ) -> self.types.GetPromptResult:  # type: ignore
            """
            Get a specific prompt with filled-in arguments.
            """
            client = await run_sync(self._get_or_create_client)

            endpoint_name = None
            for endpoint_name in self.tool_to_endpoint.values():
                block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
                if (
                    block_fn
                    and block_fn.fn
                    and hasattr(block_fn.fn, "_mcp_type")
                    and block_fn.fn._mcp_type == "prompt"
                    and block_fn.fn._mcp_name == name  # type: ignore
                ):
                    break

            if not endpoint_name:
                raise ValueError(f"Prompt not found: {name}")

            arguments = arguments or {}

            block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
            assert block_fn is not None  # noqa: S101

            if endpoint_name in self.api_info["named_endpoints"]:
                parameters_info = self.api_info["named_endpoints"][endpoint_name][
                    "parameters"
                ]
                processed_args = client_utils.construct_args(
                    parameters_info,
                    (),
                    arguments,
                )
            else:
                processed_args = list(arguments.values())

            async for update in client.submit(*processed_args, api_name=endpoint_name):
                if update.type == "output" and update.final:  # type: ignore
                    output = update.outputs  # type: ignore
                    result = output["data"][0]
                    break

            return self.types.GetPromptResult(  # type: ignore
                messages=[
                    self.types.PromptMessage(  # type: ignore
                        role="user",
                        content=self.types.TextContent(type="text", text=str(result)),  # type: ignore
                    )
                ]
            )

        return server

    def launch_mcp_on_sse(self, app: Starlette, subpath: str, root_path: str) -> None:
        """
        Launch the MCP server on the SSE transport.

        Parameters:
            app: The Gradio app to mount the MCP server on.
            subpath: The subpath to mount the MCP server on. E.g. "/gradio_api/mcp"
            root_path: The root path of the Gradio Blocks app.
        """
        messages_path = "/messages/"
        sse = self.SseServerTransport(messages_path)  # type: ignore
        self.root_path = root_path

        async def handle_sse(request):
            try:
                async with sse.connect_sse(
                    request.scope, request.receive, request._send
                ) as streams:
                    await self.mcp_server.run(
                        streams[0],
                        streams[1],
                        self.mcp_server.create_initialization_options(),
                    )
                return Response()
            except Exception as e:
                print(f"MCP SSE connection error: {str(e)}")
                raise

        app.mount(
            subpath,
            Starlette(
                routes=[
                    Route(
                        "/schema",
                        endpoint=self.get_complete_schema,  # Not required for MCP but used by the Hugging Face MCP server to get the schema for MCP Spaces without needing to establish an SSE connection
                    ),
                    Route("/sse", endpoint=handle_sse),
                    Mount("/messages/", app=sse.handle_post_message),
                    Mount("/", app=self.handle_streamable_http),
                ],
            ),
        )

    def get_block_fn_from_endpoint_name(
        self, endpoint_name: str
    ) -> "BlockFunction | None":
        """
        Get the BlockFunction for a given endpoint name (e.g. "/predict").

        Parameters:
            endpoint_name: The name of the endpoint to get the BlockFunction for.

        Returns:
            The BlockFunction for the given endpoint name, or None if it is not found.
        """
        block_fn = next(
            (
                fn
                for fn in self.blocks.fns.values()
                if fn.api_name == endpoint_name.lstrip("/")
            ),
            None,
        )
        return block_fn

    @property
    def _file_data_tool_description(self) -> str:
        """
        Sentence prompting the agent to use the upload_file_to_gradio tool if a file is passed as an input.
        """
        return " If a user passes a file as an input, use the upload_file_to_gradio tool, if present, to upload the file to the gradio app and create a Gradio File Input. Then use the returned path as the input to the tool"

    def get_fn_description(
        self, block_fn: "BlockFunction", tool_name: str
    ) -> tuple[str, dict[str, str]]:
        """
        Get the description of a function, which is used to describe the tool in the MCP server.
        Also returns the description of each parameter of the function as a dictionary.
        """
        description, parameters, returns = utils.get_function_description(block_fn.fn)  # type: ignore
        _, filedata_positions = self.get_input_schema(tool_name, parameters)
        if block_fn.api_description is False:
            description = ""
        elif block_fn.api_description is None:
            if len(filedata_positions) > 0:
                description += self._file_data_tool_description
            if returns:
                description += (
                    ("" if description.endswith(".") else ".")
                    + " Returns: "
                    + ", ".join(returns)
                )
        else:
            description = block_fn.api_description
            if len(filedata_positions) > 0:
                description += self._file_data_tool_description  # type: ignore
        assert isinstance(description, str)  # noqa: S101
        return description, parameters

    @staticmethod
    def insert_empty_state(
        inputs: Sequence["Component | BlockContext"], data: list
    ) -> list:
        """
        Insert None placeholder values for any State input components, as State inputs
        are not included in the endpoint schema.
        """
        for i, input_component_type in enumerate(inputs):
            if isinstance(input_component_type, State):
                data.insert(i, None)
        return data

    @staticmethod
    def pop_returned_state(
        components: Sequence["Component | BlockContext"], data: Any
    ) -> list:
        """
        Remove any values corresponding to State output components from the data
        as State outputs are not included in the endpoint schema.
        """
        for i, component_type in enumerate(components):
            if isinstance(component_type, State):
                data.pop(i)
        return data

    def get_input_schema(
        self,
        tool_name: str,
        parameters: dict[str, str] | None = None,
    ) -> tuple[dict[str, Any], list[list[str | int]]]:
        """
        Get the input schema of the Gradio app API, appropriately formatted for MCP.

        Parameters:
            tool_name: The name of the tool to get the schema for, e.g. "predict"
            parameters: The description and parameters of the tool to get the schema for.
        Returns:
            - The input schema of the Gradio app API.
            - A list of positions of FileData objects in the input schema.
        """
        endpoint_name = self.tool_to_endpoint.get(tool_name)
        if endpoint_name is None:
            raise ValueError(f"Unknown tool for this Gradio app: {tool_name}")
        named_endpoints = self.api_info["named_endpoints"]
        endpoint_info = named_endpoints.get(endpoint_name)
        assert endpoint_info is not None  # noqa: S101

        schema = {
            "type": "object",
            "properties": {
                p["parameter_name"]: {
                    **p["type"],
                    **(
                        {"description": parameters[p["parameter_name"]]}
                        if parameters and p["parameter_name"] in parameters
                        else {}
                    ),
                    **(
                        {"default": p["parameter_default"]}
                        if "parameter_default" in p and p["parameter_default"]
                        else {}
                    ),
                }
                for p in endpoint_info["parameters"]
            },
        }
        return self.simplify_filedata_schema(schema)

    async def get_complete_schema(self, request) -> JSONResponse:
        """
        Get the complete schema of the Gradio app API. For debugging purposes, also used by
        the Hugging Face MCP server to get the schema for MCP Spaces without needing to
        establish an SSE connection.

        Parameters:
            request: The Starlette request object.

        Returns:
            A JSONResponse containing a dictionary mapping tool names to their input schemas.
        """
        if not self.api_info:
            return JSONResponse({})

        query_params = dict(getattr(request, "query_params", {}))
        selected_tools = None
        if "tools" in query_params:
            tools = query_params["tools"].split(",")
            selected_tools = set(tools)

        file_data_present = False

        schemas = []
        for tool_name, endpoint_name in self.tool_to_endpoint.items():
            if selected_tools is not None and tool_name not in selected_tools:
                continue
            block_fn = self.get_block_fn_from_endpoint_name(endpoint_name)
            assert block_fn is not None and block_fn.fn is not None  # noqa: S101

            description, parameters = self.get_fn_description(block_fn, tool_name)
            schema, filedata_positions = self.get_input_schema(tool_name, parameters)
            if len(filedata_positions) > 0 and not file_data_present:
                file_data_present = True

            type_hints = utils.get_type_hints(block_fn.fn)
            required_headers = []
            for param_name, type_hint in type_hints.items():
                if type_hint is Header or type_hint is Optional[Header]:
                    header_name = param_name.replace("_", "-").lower()
                    required_headers.append(header_name)

            mcp_type = "tool"  # Default
            if hasattr(block_fn.fn, "_mcp_type"):
                mcp_type = block_fn.fn._mcp_type

            meta = {
                "file_data_present": file_data_present,
                "mcp_type": mcp_type,
                "endpoint_name": block_fn.api_name,
            }
            if required_headers:
                meta["headers"] = required_headers

            info = {
                "name": tool_name,
                "description": description,
                "inputSchema": schema,
                "meta": meta,
            }
            schemas.append(info)

        return JSONResponse(schemas)

    def simplify_filedata_schema(
        self, schema: dict[str, Any]
    ) -> tuple[dict[str, Any], list[list[str | int]]]:
        """
        Parses a schema of a Gradio app API to identify positions of FileData objects. Replaces them with base64
        strings while keeping track of their positions so that they can be converted back to FileData objects
        later.

        Parameters:
            schema: The original schema of the Gradio app API.

        Returns:
            A tuple containing the simplified schema and the positions of the FileData objects.
        """

        def is_gradio_filedata(obj: Any, defs: dict[str, Any]) -> bool:
            if not isinstance(obj, dict):
                return False

            if "$ref" in obj:
                ref = obj["$ref"]
                if ref.startswith("#/$defs/"):
                    key = ref.split("/")[-1]
                    obj = defs.get(key, {})
                else:
                    return False

            props = obj.get("properties", {})
            meta = props.get("meta", {})

            if "$ref" in meta:
                ref = meta["$ref"]
                if ref.startswith("#/$defs/"):
                    key = ref.split("/")[-1]
                    meta = defs.get(key, {})
                else:
                    return False

            type_field = meta.get("properties", {}).get("_type", {})
            default_type = meta.get("default", {}).get("_type")
            return (
                type_field.get("const") == "gradio.FileData"
                or default_type == "gradio.FileData"
            )

        def traverse(
            node: Any,
            path: list[str | int] | None = None,
            defs: dict[str, Any] | None = None,
        ) -> Any:
            if path is None:
                path = []
            if defs is None:
                defs = {}
            # Deep copy the node to avoid modifying the original node
            node = copy.deepcopy(node)

            if isinstance(node, dict):
                if "$defs" in node:
                    defs.update(node["$defs"])

                if is_gradio_filedata(node, defs):
                    filedata_positions.append(path.copy())
                    for key in ["properties", "additional_description", "$defs"]:
                        node.pop(key, None)
                    node["type"] = "string"
                    node["format"] = "Gradio File Input - a http or https url to a file"

                result = {}
                is_schema_root = "type" in node and "properties" in node
                for key, value in node.items():
                    if is_schema_root and key == "properties":
                        result[key] = traverse(value, path, defs)
                    else:
                        path.append(key)
                        result[key] = traverse(value, path, defs)
                        path.pop()
                return result

            elif isinstance(node, list):
                result = []
                for i, item in enumerate(node):
                    path.append(i)
                    result.append(traverse(item, path, defs))
                    path.pop()
                return result

            return node

        filedata_positions: list[list[str | int]] = []
        simplified_schema = traverse(schema)
        return simplified_schema, filedata_positions

    def convert_strings_to_filedata(
        self, value: Any, filedata_positions: list[list[str | int]]
    ) -> Any:
        """
        Convert specific string values back to FileData objects based on their positions.
        This is used to convert string values (as base64 encoded strings) to FileData
        dictionaries so that they can be passed into .preprocess() logic of a Gradio app.

        Parameters:
            value: The input data to process, which can be an arbitrary nested data structure
                that may or may not contain strings that should be converted to FileData objects.
            filedata_positions: List of paths to positions in the input data that should be converted to FileData objects.

        Returns:
            The processed data with strings converted to FileData objects where appropriate. Base64
            encoded strings are first saved to a temporary file and then converted to a FileData object.

        Example:
            >>> convert_strings_to_filedata(
                {"image": "data:image/jpeg;base64,..."},
                [["image"]]
            )
            >>> {'image': FileData(path='<temporary file path>')},
        """

        def traverse(node: Any, path: list[str | int] | None = None) -> Any:
            if path is None:
                path = []

            if isinstance(node, dict):
                return {
                    key: traverse(value, path + [key]) for key, value in node.items()
                }
            elif isinstance(node, list):
                return [traverse(item, path + [i]) for i, item in enumerate(node)]
            elif isinstance(node, str) and path in filedata_positions:
                if node.startswith("data:"):
                    # Even though base64 is not officially part of our schema, some MCP clients
                    # might return base64 encoded strings, so try to save it to a temporary file.
                    return handle_file(
                        processing_utils.save_base64_to_cache(node, DEFAULT_TEMP_DIR)
                    )
                elif node.startswith(("http://", "https://")):
                    return handle_file(node)
                else:
                    raise ValueError(
                        f"Invalid file data format, provide a url ('http://...' or 'https://...'). Received: {node}"
                    )
            return node

        return traverse(value)

    @staticmethod
    def get_image(file_path: str) -> Image.Image | None:
        """
        If a filepath is a valid image, returns a PIL Image object. Otherwise returns None.
        """
        if not os.path.exists(file_path):
            return None
        ext = os.path.splitext(file_path.lower())[1]
        if ext not in Image.registered_extensions():
            return None
        try:
            return Image.open(file_path)
        except Exception:
            return None

    @staticmethod
    def get_svg(file_data: Any) -> bytes | None:
        """
        If a file_data is a valid FileDataDict with a url that is a data:image/svg+xml, returns bytes of the svg. Otherwise returns None.
        """
        if isinstance(file_data, dict) and (url := file_data.get("url")):
            if isinstance(url, str) and url.startswith("data:image/svg"):
                return unquote(url.split(",", 1)[1]).encode()
            else:
                return None
        else:
            return None

    @staticmethod
    def get_base64_data(image: Image.Image, format: str) -> str:
        """
        Returns a base64 encoded string of the image.
        """
        buffer = BytesIO()
        image.save(buffer, format=format)
        return base64.b64encode(buffer.getvalue()).decode("utf-8")

    def postprocess_output_data(
        self, data: Any, root_url: str
    ) -> list["types.TextContent | types.ImageContent"]:
        """
        Postprocess the output data from the Gradio app to convert FileData objects back to base64 encoded strings.

        Parameters:
            data: The output data to postprocess.
        """
        return_values = []
        data = processing_utils.add_root_url(data, root_url, None)
        for output in data:
            if svg_bytes := self.get_svg(output):
                base64_data = base64.b64encode(svg_bytes).decode("utf-8")
                mimetype = "image/svg+xml"
                svg_path = processing_utils.save_bytes_to_cache(
                    svg_bytes, f"{output['orig_name']}", DEFAULT_TEMP_DIR
                )
                svg_url = f"{root_url}/gradio_api/file={svg_path}"
                return_value = [
                    self.types.ImageContent(  # type: ignore
                        type="image", data=base64_data, mimeType=mimetype
                    ),
                    self.types.TextContent(  # type: ignore
                        type="text",
                        text=f"SVG Image URL: {svg_url}",
                    ),
                ]
            elif client_utils.is_file_obj_with_meta(output):
                if image := self.get_image(output["path"]):
                    image_format = image.format or "png"
                    base64_data = self.get_base64_data(image, image_format)
                    mimetype = f"image/{image_format.lower()}"
                    return_value = [
                        self.types.ImageContent(  # type: ignore
                            type="image", data=base64_data, mimeType=mimetype
                        ),
                        self.types.TextContent(  # type: ignore
                            type="text",
                            text=f"Image URL: {output['url'] or output['path']}",
                        ),
                    ]
                else:
                    return_value = [
                        self.types.TextContent(  # type: ignore
                            type="text", text=str(output["url"] or output["path"])
                        )
                    ]
            else:
                return_value = [self.types.TextContent(type="text", text=str(output))]  # type: ignore
            return_values.extend(return_value)
        return return_values


######################################################
### MCP decorators that add metadata to functions.
######################################################


def resource(
    uri_template: str, description: str | None = None, mime_type: str | None = None
):
    """Decorator to mark a function as an MCP resource."""

    def decorator(fn):
        fn._mcp_type = "resource"
        fn._mcp_uri_template = uri_template
        fn._mcp_description = description
        fn._mcp_mime_type = mime_type or "text/plain"
        return fn

    return decorator


def prompt(name: str | None = None, description: str | None = None):
    """Decorator to mark a function as an MCP prompt."""

    def decorator(fn):
        fn._mcp_type = "prompt"
        fn._mcp_name = name or fn.__name__
        fn._mcp_description = description
        return fn

    return decorator


def tool(
    name: str | None = None,
    description: str | None = None,
    structured_output: bool = False,
    _meta: dict[str, Any] | None = None,
):
    """
    Decorator to mark a function as an MCP tool (optional, since functions are registered as tools by default).
    Can be used to configure various aspects of the tool.

    Parameters:
        name: The name of the tool. Overrides the default name of the function.
        description: The description of the tool. Overrides the default description from the function's docstring.
        structured_output: Whether the tool should return structured output (implementation is quite limited at the moment). If True, the output will be wrapped in a dictionary with the key "result" and the value being the output of the function. Recommended to keep this False unless you have a specific reason to need the structured output.
        _meta: Additional metadata for the tool.
    """

    def decorator(fn):
        fn._mcp_type = "tool"
        fn._mcp_name = name
        fn._mcp_structured_output = structured_output
        fn._mcp_description = description
        fn._mcp_meta = _meta
        return fn

    return decorator



import copy
import json
import os
import tempfile

import pytest
from mcp import ClientSession
from mcp.client.streamable_http import streamablehttp_client
from PIL import Image
from starlette.requests import Request

import gradio as gr
from gradio.data_classes import FileData
from gradio.mcp import GradioMCPServer


def test_gradio_mcp_server_initialization(test_mcp_app):
    server = GradioMCPServer(test_mcp_app)
    assert server.blocks == test_mcp_app
    assert server.mcp_server is not None


def test_get_block_fn_from_tool_name(test_mcp_app):
    server = GradioMCPServer(test_mcp_app)
    result = server.get_block_fn_from_endpoint_name("test_tool")
    assert result == test_mcp_app.fns[0]
    result = server.get_block_fn_from_endpoint_name("nonexistent_tool")
    assert result is None


def test_generate_tool_names_correctly_for_interfaces():
    def echo(x):
        return x

    class MyCallable:
        def __call__(self, x):
            return x

    app = gr.TabbedInterface(
        [
            gr.Interface(echo, "text", "text"),
            gr.Interface(echo, "image", "image"),
            gr.Interface(lambda x: x, "audio", "audio"),
            gr.Interface(MyCallable(), "text", "text"),
        ]
    )
    server = GradioMCPServer(app)
    assert list(server.tool_to_endpoint.keys()) == [
        "echo",
        "echo_1",
        "_lambda_",
        "MyCallable",
    ]


def test_convert_strings_to_filedata(test_mcp_app):
    server = GradioMCPServer(test_mcp_app)

    test_data = {
        "text": "test text",
        "image": "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mP8z8BQDwAEhQGAhKmMIQAAAABJRU5ErkJggg==",
    }
    filedata_positions: list[list[str | int]] = [["image"]]
    result = server.convert_strings_to_filedata(test_data, filedata_positions)
    assert isinstance(result["image"], dict)
    result["image"] = FileData(**result["image"])  # type: ignore
    assert os.path.exists(result["image"].path)

    test_data = {"image": "invalid_data"}
    with pytest.raises(ValueError):
        server.convert_strings_to_filedata(test_data, filedata_positions)


def test_postprocess_output_data(test_mcp_app):
    server = GradioMCPServer(test_mcp_app)
    fake_root_url = "http://localhost:7860"

    temp_file = tempfile.NamedTemporaryFile(suffix=".png", delete=False)
    temp_file.close()
    try:
        img = Image.new("RGB", (10, 10), color="red")
        img.save(temp_file.name)
        url = f"http://localhost:7860/gradio_api/file={temp_file.name}"
        test_data = [
            {"path": temp_file.name, "url": url, "meta": {"_type": "gradio.FileData"}}
        ]
        result = server.postprocess_output_data(test_data, fake_root_url)
        assert len(result) == 2
        assert result[0].type == "image"
        assert result[0].mimeType == "image/png"  # type: ignore
        assert result[1].type == "text"
        assert url in result[1].text  # type: ignore
    finally:
        os.unlink(temp_file.name)

    svg_data_uri = "data:image/svg+xml,%3Csvg%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%20width%3D%22100%22%20height%3D%22100%22%3E%3Ccircle%20cx%3D%2250%22%20cy%3D%2250%22%20r%3D%2240%22%20fill%3D%22blue%22%2F%3E%3C%2Fsvg%3E"
    test_data = [
        {
            "path": None,
            "url": svg_data_uri,
            "meta": {"_type": "gradio.FileData"},
            "orig_name": "test.svg",
        }
    ]
    result = server.postprocess_output_data(test_data, fake_root_url)
    assert len(result) == 2
    assert result[0].type == "image"
    assert result[0].mimeType == "image/svg+xml"  # type: ignore
    assert result[1].type == "text"
    assert "Image URL:" in result[1].text  # type: ignore
    assert "/gradio_api/file=" in result[1].text  # type: ignore

    test_data = ["test text"]
    result = server.postprocess_output_data(test_data, fake_root_url)
    assert len(result) == 1
    assert result[0].type == "text"
    assert result[0].text == "test text"  # type: ignore


def test_simplify_filedata_schema(test_mcp_app):
    server = GradioMCPServer(test_mcp_app)

    test_schema = {
        "type": "object",
        "properties": {
            "text": {"type": "string"},
            "image": {
                "type": "object",
                "properties": {"meta": {"default": {"_type": "gradio.FileData"}}},
            },
        },
    }
    old_schema = copy.deepcopy(test_schema)

    simplified_schema, filedata_positions = server.simplify_filedata_schema(test_schema)
    assert simplified_schema["properties"]["image"]["type"] == "string"
    assert filedata_positions == [["image"]]
    # Verify that the original schema is not modified
    assert test_schema == old_schema


def test_tool_prefix_character_replacement(test_mcp_app):
    test_cases = [
        ("test-space", "test_space_test_tool"),
        ("flux.1_schnell", "flux_1_schnell_test_tool"),
        ("test\\backslash", "test_backslash_test_tool"),
        ("test:colon spaces ", "test_colon_spaces__test_tool"),
    ]

    original_system = os.environ.get("SYSTEM")
    original_space_id = os.environ.get("SPACE_ID")

    try:
        os.environ["SYSTEM"] = "spaces"
        for space_id, tool_name in test_cases:
            os.environ["SPACE_ID"] = space_id
            server = GradioMCPServer(test_mcp_app)
            assert tool_name in server.tool_to_endpoint
    finally:
        if original_system is not None:
            os.environ["SYSTEM"] = original_system
        else:
            os.environ.pop("SYSTEM", None)
        if original_space_id is not None:
            os.environ["SPACE_ID"] = original_space_id
        else:
            os.environ.pop("SPACE_ID", None)


@pytest.mark.asyncio
async def test_associative_keyword_in_schema():
    def test_tool(x):
        return x

    demo = gr.Interface(test_tool, "image", "image")
    server = GradioMCPServer(demo)

    scope = {
        "type": "http",
        "headers": [],
        "server": ("localhost", 7860),
        "path": "/gradio_api/mcp/schema",
        "query_string": "",
    }
    request = Request(scope)

    schema = (await server.get_complete_schema(request)).body.decode("utf-8")  # type: ignore
    schema = json.loads(schema)
    assert (
        schema[0]["inputSchema"]["properties"]["x"]["format"]
        == "Gradio File Input - a http or https url to a file"
    )
    assert (
        "to upload the file to the gradio app and create a Gradio File Input"
        in schema[0]["description"]
    )
    assert schema[0]["meta"]["file_data_present"]


@pytest.mark.asyncio
async def test_tool_selection_via_query_params():
    def tool_1(x):
        return x

    def tool_2(x):
        return x

    demo = gr.TabbedInterface(
        [
            gr.Interface(tool_1, "image", "image"),
            gr.Interface(tool_2, "image", "image"),
        ]
    )

    server = GradioMCPServer(demo)

    scope = {
        "type": "http",
        "headers": [],
        "server": ("localhost", 7860),
        "path": "/gradio_api/mcp/schema",
        "query_string": "",
    }
    request = Request(scope)

    schema = (await server.get_complete_schema(request)).body.decode("utf-8")  # type: ignore
    schema = json.loads(schema)
    assert schema[0]["name"] == "tool_1"
    assert schema[1]["name"] == "tool_2"

    scope = {
        "type": "http",
        "headers": [],
        "server": ("localhost", 7860),
        "path": "/gradio_api/mcp/schema",
        "query_string": "tools=tool_2",
    }
    request = Request(scope)

    schema = (await server.get_complete_schema(request)).body.decode("utf-8")  # type: ignore
    schema = json.loads(schema)
    assert len(schema) == 1
    assert schema[0]["name"] == "tool_2"


@pytest.mark.asyncio
@pytest.mark.serial
async def test_mcp_streamable_http_client():
    def double(word: str) -> str:
        """
        Doubles the input word.

        Parameters:
            word: The word to double
        Returns:
            The doubled word
        """
        return word * 2

    with gr.Blocks() as demo:
        input_box = gr.Textbox(label="Input")
        output_box = gr.Textbox(label="Output")
        input_box.change(double, input_box, output_box, api_name="double")

    _, local_url, _ = demo.launch(prevent_thread_lock=True, mcp_server=True)
    mcp_url = f"{local_url}gradio_api/mcp/"

    try:
        async with streamablehttp_client(mcp_url) as (read_stream, write_stream, _):
            async with ClientSession(read_stream, write_stream) as session:
                await session.initialize()

                tools_response = await session.list_tools()
                assert len(tools_response.tools) == 1
                tool = tools_response.tools[0]
                assert "double" in tool.name
                assert "Doubles the input word" in tool.description  # type: ignore

                result = await session.call_tool(tool.name, arguments={"word": "Hello"})  # type: ignore
                assert len(result.content) == 1  # type: ignore
                assert result.content[0].text == "HelloHello"  # type: ignore
    finally:
        demo.close()


@pytest.mark.serial
@pytest.mark.asyncio
async def test_mcp_streamable_http_client_with_progress_callback():
    progress_updates = []

    def slow_processor(text: str, progress=gr.Progress()) -> str:
        """
        Processes text slowly with progress updates.

        Parameters:
            text: The text to process
        Returns:
            The processed text
        """
        total = len(text)
        for i in range(total):
            progress((i + 1) / total, desc=f"Processing character {i + 1}/{total}")
        return text.upper()

    with gr.Blocks() as demo:
        input_box = gr.Textbox(label="Input")
        output_box = gr.Textbox(label="Output")
        input_box.submit(slow_processor, input_box, output_box, api_name="process")

    demo.queue()
    _, local_url, _ = demo.launch(prevent_thread_lock=True, mcp_server=True)
    mcp_url = f"{local_url}gradio_api/mcp/"

    try:
        async with streamablehttp_client(mcp_url) as (read_stream, write_stream, _):
            async with ClientSession(read_stream, write_stream) as session:
                await session.initialize()

                tools_response = await session.list_tools()
                assert len(tools_response.tools) == 1
                tool = tools_response.tools[0]
                assert "process" in tool.name

                async def progress_callback(
                    progress: float, total: float | None, message: str | None
                ):
                    progress_updates.append(
                        {"progress": progress, "total": total, "message": message}
                    )

                result = await session.call_tool(
                    tool.name,
                    arguments={"text": "test"},
                    progress_callback=progress_callback,
                    meta={"progressToken": "test-token-123"},
                )

                assert len(result.content) == 1  # type: ignore
                assert result.content[0].text == "TEST"  # type: ignore
                assert len(progress_updates) > 0, "Expected to receive progress updates"
    finally:
        demo.close()


@pytest.mark.asyncio
@pytest.mark.serial
async def test_mcp_streamable_http_client_with_stateful_app(stateful_mcp_app):
    _, local_url, _ = stateful_mcp_app.launch(prevent_thread_lock=True, mcp_server=True)
    mcp_url = f"{local_url}gradio_api/mcp/"

    try:
        async with streamablehttp_client(mcp_url) as (read_stream, write_stream, _):
            async with ClientSession(read_stream, write_stream) as session:
                await session.initialize()

                tools_response = await session.list_tools()
                assert len(tools_response.tools) == 1
                tool = tools_response.tools[0]

                result = await session.call_tool(
                    tool.name,
                    arguments={"name": "test", "flag": True, "gallery_images": 42},
                )
                assert len(result.content) == 1  # type: ignore
                assert (
                    result.content[0].text  # type: ignore
                    == "name=test, hidden_state=hidden_value, flag=True, gallery=42"
                )
    finally:
        stateful_mcp_app.close()


@pytest.mark.asyncio
@pytest.mark.serial
async def test_x_gradio_user_mcp_gets_set():
    def fn(name: str, request: gr.Request) -> str:
        return f"Hello, {name}! Your x-gradio-user is {request.headers.get('x-gradio-user', 'not provided')}"

    app = gr.Interface(fn, "text", "text")
    _, local_url, _ = app.launch(prevent_thread_lock=True, mcp_server=True)
    mcp_url = f"{local_url}gradio_api/mcp/"

    try:
        async with streamablehttp_client(mcp_url) as (read_stream, write_stream, _):
            async with ClientSession(read_stream, write_stream) as session:
                await session.initialize()

                tools_response = await session.list_tools()
                tool = tools_response.tools[0]

                result = await session.call_tool(
                    tool.name,
                    arguments={"name": "Gradio"},
                )
                assert len(result.content) == 1  # type: ignore
                assert (
                    result.content[0].text  # type: ignore
                    == "Hello, Gradio! Your x-gradio-user is mcp"
                )
    finally:
        app.close()



"""gr.Navbar() component."""

from __future__ import annotations

from typing import Any, Literal

from gradio_client.documentation import document

from gradio.components.base import Component
from gradio.events import Events


@document()
class Navbar(Component):
    """
    Creates a navigation bar component for multipage Gradio apps. The navbar component allows customizing the
    appearance of the navbar for that page. Only one Navbar component can exist per page in a Blocks app,
    and it can be placed anywhere within the page.

    The Navbar component is designed to control the appearance of the navigation bar in multipage
    applications. When present in a Blocks app, its properties override the default navbar behavior.

    Example:
        ```python
        import gradio as gr

        with gr.Blocks() as demo:
            navbar = gr.Navbar(
                visible=True,
                main_page_name="My App",
                value=[("Analytics", "analytics"), ("About", "https://twitter.com/abidlabs")]
            )
            gr.Textbox(label="Main page content")

        with demo.route("About"):
            gr.Markdown("This is the about page")

        demo.launch()
        ```
    """

    EVENTS = [Events.change]

    def __init__(
        self,
        value: list[tuple[str, str]] | None = None,
        *,
        visible: bool = True,
        main_page_name: str | Literal[False] = "Home",
        elem_id: str | None = None,
        elem_classes: list[str] | str | None = None,
        render: bool = True,
        key: int | str | tuple[int | str, ...] | None = None,
    ):
        """
        Parameters:
            value: If a list of tuples of (page_name, page_path) are provided, these additional pages will be added to the navbar alongside the existing pages defined in the Blocks app. The page_path can be either a relative path for internal Gradio app pages (e.g., "analytics") or an absolute URL for external links (e.g., "https://twitter.com/username"). Otherwise, only the pages defined using the `Blocks.route` method will be displayed. Example: [("Dashboard", "dashboard"), ("About", "https://twitter.com/abidlabs")]
            visible: If True, the navbar will be visible. If False, the navbar will be hidden.
            main_page_name: The title to display in the navbar for the main page of the Gradio. If False, the main page will not be displayed in the navbar.
            elem_id: An optional string that is assigned as the id of this component in the HTML DOM. Can be used for targeting CSS styles.
            elem_classes: An optional list of strings that are assigned as the classes of this component in the HTML DOM. Can be used for targeting CSS styles.
            render: If False, component will not render be rendered in the Blocks context. Should be used if the intention is to assign event listeners now but render the component later.
            key: in a gr.render, Components with the same key across re-renders are treated as the same component, not a new component.
        """
        self.visible = visible
        self.main_page_name = main_page_name

        super().__init__(
            elem_id=elem_id,
            elem_classes=elem_classes,
            render=render,
            key=key,
            visible=visible,
            value=value,
        )

    def preprocess(
        self, payload: list[tuple[str, str]] | None
    ) -> list[tuple[str, str]] | None:
        return payload

    def postprocess(
        self, value: list[tuple[str, str]] | None
    ) -> list[tuple[str, str]] | None:
        return value

    def api_info(self) -> dict[str, Any]:
        return {}

    def example_payload(self) -> list[tuple[str, str]] | None:
        return None

    def example_value(self) -> list[tuple[str, str]] | None:
        return None



import numpy as np
import PIL

import gradio as gr


class TestAnnotatedImage:
    def test_postprocess(self):
        """
        postprocess
        """
        component = gr.AnnotatedImage()
        img = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
        mask1 = [40, 40, 50, 50]
        mask2 = np.zeros((100, 100), dtype=np.uint8)
        mask2[10:20, 10:20] = 1

        input = (img, [(mask1, "mask1"), (mask2, "mask2")])
        assert (result := component.postprocess(input))  # type: ignore
        result = result.model_dump()

        base_img_out = PIL.Image.open(result["image"]["path"])  # type: ignore

        assert result["annotations"][0]["label"] == "mask1"

        mask1_img_out = PIL.Image.open(result["annotations"][0]["image"]["path"])  # type: ignore
        assert mask1_img_out.size == base_img_out.size
        mask1_array_out = np.array(mask1_img_out)
        assert np.max(mask1_array_out[40:50, 40:50]) == 255
        assert np.max(mask1_array_out[50:60, 50:60]) == 0

    def test_annotated_image_format_parameter(self):
        component = gr.AnnotatedImage(format="jpeg")
        img = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
        mask1 = (40, 40, 50, 50)
        data = (img, [(mask1, "mask1"), (mask1, "mask2")])
        assert (output := component.postprocess(data))
        assert output.image.path.endswith(".jpeg")
        assert output.annotations[0].image.path.endswith(".png")

    def test_component_functions(self):
        ht_output = gr.AnnotatedImage(label="sections", show_legend=False)
        assert ht_output.get_config() == {
            "name": "annotatedimage",
            "show_label": True,
            "label": "sections",
            "show_legend": False,
            "container": True,
            "min_width": 160,
            "scale": None,
            "format": "webp",
            "color_map": None,
            "height": None,
            "width": None,
            "elem_id": None,
            "elem_classes": [],
            "visible": True,
            "value": None,
            "proxy_url": None,
            "_selectable": False,
            "key": None,
            "preserved_by_key": ["value"],
            "buttons": ["fullscreen"],
        }



from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Literal, Union, cast

from gradio_client.documentation import document

from gradio.blocks import Block
from gradio.components import Component
from gradio.context import Context, LocalContext
from gradio.events import EventListener, EventListenerMethod
from gradio.layouts import Column, Row

if TYPE_CHECKING:
    from gradio.blocks import BlockFunction
    from gradio.events import EventListenerCallable


class Renderable:
    def __init__(
        self,
        fn: Callable,
        inputs: Sequence[Component],
        triggers: list[tuple[Block | None, str]],
        concurrency_limit: int | None | Literal["default"],
        concurrency_id: str | None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None,
        queue: bool,
        show_progress: Literal["full", "minimal", "hidden"],
    ):
        if Context.root_block is None:
            raise ValueError("Reactive render must be inside a Blocks context.")

        self._id = len(Context.root_block.renderables)
        Context.root_block.renderables.append(self)
        self.ContainerClass = Row if isinstance(Context.block, Row) else Column
        self.container = self.ContainerClass(show_progress=True)
        self.container_id = self.container._id

        self.fn = fn
        self.inputs = inputs
        self.triggers: list[EventListenerMethod] = []
        self.page = Context.root_block.current_page
        self.key_to_id_map: dict[int | str | tuple[str | int, ...], int] = {}
        self.render_iteration = 0

        self.triggers = [EventListenerMethod(*t) for t in triggers]
        Context.root_block.default_config.set_event_trigger(
            self.triggers,
            self.apply,
            self.inputs,
            self.container,
            api_visibility="undocumented",
            concurrency_limit=concurrency_limit,
            concurrency_id=concurrency_id,
            renderable=self,
            trigger_mode=trigger_mode,
            postprocess=False,
            queue=queue,
            show_progress=show_progress,
        )

    def apply(self, *args, **kwargs):
        blocks_config = LocalContext.blocks_config.get(None)
        if blocks_config is None:
            raise ValueError("Reactive render must be inside a LocalContext.")

        fns_from_last_render: list[BlockFunction] = []
        for fn in blocks_config.fns.values():
            if fn.rendered_in is self:
                fns_from_last_render.append(fn)

        container_copy = self.ContainerClass(render=False, show_progress=True)
        container_copy._id = self.container_id
        container_copy.page = self.page
        LocalContext.renderable.set(self)
        LocalContext.key_to_id_map.set(self.key_to_id_map)

        try:
            with container_copy:
                self.fn(*args, **kwargs)
                blocks_config.blocks[self.container_id] = container_copy
                blocks_config.attach_load_events(self)
        finally:
            LocalContext.renderable.set(None)
            LocalContext.key_to_id_map.set(None)

        for fn in fns_from_last_render:
            if blocks_config.fns[fn._id].render_iteration != self.render_iteration:
                del blocks_config.fns[fn._id]

        self.render_iteration += 1


@document()
def render(
    inputs: Sequence[Component] | Component | None = None,
    triggers: Sequence[EventListenerCallable] | EventListenerCallable | None = None,
    *,
    queue: bool = True,
    trigger_mode: Literal["once", "multiple", "always_last"] | None = "always_last",
    concurrency_limit: int | None | Literal["default"] = None,
    concurrency_id: str | None = None,
    show_progress: Literal["full", "minimal", "hidden"] = "full",
):
    """
    The render decorator allows Gradio Blocks apps to have dynamic layouts, so that the components and event listeners in your app can change depending on custom logic.
    Attaching a @gr.render decorator to a function will cause the function to be re-run whenever the inputs are changed (or specified triggers are activated). The function contains the components and event listeners that will update based on the inputs.

    The basic usage of @gr.render is as follows:\n
    1. Create a function and attach the @gr.render decorator to it.\n
    2. Add the input components to the `inputs=` argument of @gr.render, and create a corresponding argument in your function for each component.\n
    3. Add all components inside the function that you want to update based on the inputs. Any event listeners that use these components should also be inside this function.\n
    Parameters:
        inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
        triggers: List of triggers to listen to, e.g. [btn.click, number.change]. If None, will listen to changes to any inputs.
        queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
        trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` and `.key_up()` events) would allow a second submission after the pending event is complete.
        concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
        concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
    Example:
        import gradio as gr

        with gr.Blocks() as demo:
            input_text = gr.Textbox()

            @gr.render(inputs=input_text)
            def show_split(text):
                if len(text) == 0:
                    gr.Markdown("## No Input Provided")
                else:
                    for letter in text:
                        with gr.Row():
                            text = gr.Textbox(letter)
                            btn = gr.Button("Clear")
                            btn.click(lambda: gr.Textbox(value=""), None, text)
    """
    new_triggers = cast(Union[list[EventListener], EventListener, None], triggers)

    if Context.root_block is None:
        raise ValueError("Reactive render must be inside a Blocks context.")

    inputs = (
        [inputs] if isinstance(inputs, Component) else [] if inputs is None else inputs
    )
    _triggers: list[tuple[Block | None, str]] = []
    if new_triggers is None:
        _triggers = [(Context.root_block, "load")]
        for input in inputs:
            if hasattr(input, "change"):
                _triggers.append((input, "change"))
    else:
        new_triggers = (
            [new_triggers] if isinstance(new_triggers, EventListener) else new_triggers
        )
        _triggers = [
            (getattr(t, "__self__", None) if t.has_trigger else None, t.event_name)
            for t in new_triggers
        ]

    if new_triggers:
        for trigger in new_triggers:
            if trigger.callback:
                trigger.callback(trigger.__self__)  # type: ignore

    def wrapper_function(fn):
        Renderable(
            fn,
            inputs,  # type: ignore
            _triggers,
            concurrency_limit,
            concurrency_id,
            trigger_mode,
            queue,
            show_progress,
        )
        return fn

    return wrapper_function



from __future__ import annotations

from collections.abc import Iterable

from gradio.themes.base import Base
from gradio.themes.utils import colors, fonts, sizes


class Citrus(Base):
    def __init__(
        self,
        *,
        primary_hue: colors.Color | str = colors.amber,
        secondary_hue: colors.Color | str = colors.amber,
        neutral_hue: colors.Color | str = colors.stone,
        spacing_size: sizes.Size | str = sizes.spacing_lg,
        radius_size: sizes.Size | str = sizes.radius_md,
        text_size: sizes.Size | str = sizes.text_md,
        font: fonts.Font | str | Iterable[fonts.Font | str] = (
            fonts.GoogleFont("Ubuntu"),
            "ui-sans-serif",
            "system-ui",
            "sans-serif",
        ),
        font_mono: fonts.Font | str | Iterable[fonts.Font | str] = (
            fonts.GoogleFont("Roboto Mono"),
            "ui-monospace",
            "Consolas",
            "monospace",
        ),
    ):
        super().__init__(
            primary_hue=primary_hue,
            secondary_hue=secondary_hue,
            neutral_hue=neutral_hue,
            spacing_size=spacing_size,
            radius_size=radius_size,
            text_size=text_size,
            font=font,
            font_mono=font_mono,
        )
        self.name = "citrus"
        super().set(
            slider_color="*primary_400",
            background_fill_primary="*neutral_50",
            button_primary_background_fill="*primary_500",
            button_primary_text_color="*button_secondary_text_color",
            button_secondary_background_fill="*primary_400",
            button_primary_background_fill_hover="*button_primary_background_fill",
            button_secondary_background_fill_hover="*button_secondary_background_fill",
            button_primary_shadow_hover="0px 5px 0px 0px *primary_400;",
            button_primary_shadow="0px 3px 0px 0px *primary_400;",
            button_primary_shadow_active="0px 2px 0px 0px *primary_400;",
            button_secondary_shadow_hover="0px 5px 0px 0px *primary_300;",
            button_secondary_shadow="0px 3px 0px 0px *primary_300;",
            button_secondary_shadow_active="0px 2px 0px 0px *primary_300;",
            input_shadow="0px -1px 0px 0px *neutral_300;",
            input_shadow_focus="0px -1px 0px 0px *primary_300;",
            input_background_fill="*neutral_50",
            input_background_fill_focus="*primary_50",
            block_shadow="0px 3px 0px 0px *neutral_300;",
            checkbox_label_background_fill="*neutral_200",
            checkbox_label_background_fill_hover="*checkbox_label_background_fill",
            checkbox_label_background_fill_selected="*primary_400",
            checkbox_label_border_color_selected="*primary_300",
            checkbox_label_border_width="2px",
            slider_color_dark="*primary_500",
            button_primary_background_fill_dark="*primary_600",
            button_secondary_text_color_dark="*neutral_900",
            button_primary_text_color_dark="*button_secondary_text_color",
            button_secondary_background_fill_dark="*primary_500",
            button_primary_background_fill_hover_dark="*button_primary_background_fill",
            button_secondary_background_fill_hover_dark="*button_secondary_background_fill",
            button_primary_shadow_hover_dark="0px 5px 0px 0px *primary_700;",
            button_primary_shadow_dark="0px 3px 0px 0px *primary_700;",
            button_primary_shadow_active_dark="0px 2px 0px 0px *primary_700;",
            button_secondary_shadow_hover_dark="0px 5px 0px 0px *primary_600;",
            button_secondary_shadow_dark="0px 3px 0px 0px *primary_600;",
            button_secondary_shadow_active_dark="0px 2px 0px 0px *primary_600;",
            input_shadow_dark="0px -1px 0px 0px *neutral_700;",
            input_shadow_focus_dark="0px -1px 0px 0px *primary_600;",
            input_background_fill_dark="*neutral_900",
            input_background_fill_focus_dark="none",
            block_shadow_dark="0px 3px 0px 0px *neutral_700;",
            checkbox_label_background_fill_dark="*neutral_700",
            checkbox_label_background_fill_hover_dark="*checkbox_label_background_fill",
            checkbox_label_background_fill_selected_dark="*primary_500",
            checkbox_label_border_color_selected_dark="*primary_600",
            checkbox_label_text_color_selected_dark="*button_primary_text_color",
            checkbox_background_color_selected_dark="*primary_600",
            checkbox_background_color_dark="*neutral_400",
            checkbox_label_border_width_dark="2px",
            # Stop/Cancel Button
            button_cancel_background_fill=colors.red.c500,
            button_cancel_background_fill_dark=colors.red.c700,
            button_cancel_background_fill_hover=colors.red.c600,
            button_cancel_background_fill_hover_dark=colors.red.c800,
            button_cancel_text_color="white",
            button_cancel_text_color_dark="white",
            button_cancel_text_color_hover="white",
            button_cancel_text_color_hover_dark="white",
            button_cancel_shadow="0px 3px 0px 0px rgb(248 113 113)",
            button_cancel_shadow_hover="0px 5px 0px 0px rgb(248 113 113)",
            button_cancel_shadow_active="0px 2px 0px 0px rgb(248 113 113)",
            button_cancel_shadow_dark="0px 3px 0px 0px rgb(220 38 38)",
            button_cancel_shadow_hover_dark="0px 5px 0px 0px rgb(220 38 38)",
            button_cancel_shadow_active_dark="0px 2px 0px 0px rgb(220 38 38)",
            button_transform_hover="translateY(-2px);",
            button_transform_active="translateY(1px);",
            button_transition="all 0.1s;",
            button_border_width="0px",
            panel_border_width="1px",
            block_border_width="1px",
            block_border_color="*neutral_300",
            block_background_fill="*neutral_100",
            input_border_width="1px",
            block_label_shadow="none",
            checkbox_shadow="none",
        )
